# Generated by Django 2.0 on 2018-08-08 12:14

from django.db import migrations, models
import django.db.models.deletion
import django_extensions.db.fields


class Migration(migrations.Migration):

    dependencies = [
        ('djconnectwise', '0063_merge_20180611_0905'),
    ]

    operations = [
        migrations.CreateModel(
            name='Sla',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', django_extensions.db.fields.CreationDateTimeField(auto_now_add=True, verbose_name='created')),
                ('modified', django_extensions.db.fields.ModificationDateTimeField(auto_now=True, verbose_name='modified')),
                ('name', models.TextField(max_length=250)),
                ('default_flag', models.BooleanField(default=False)),
                ('respond_hours', models.BigIntegerField()),
                ('plan_within', models.BigIntegerField()),
                ('resolution_hours', models.BigIntegerField()),
            ],
            options={
                'ordering': ('-modified', '-created'),
                'abstract': False,
                'get_latest_by': 'modified',
            },
        ),
        migrations.AddField(
            model_name='ticket',
            name='sla',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='tickets', to='djconnectwise.Sla'),
        ),
    ]



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('djconnectwise', '0019_ticket_owner'),
        ('djconnectwise', '0019_syncjob_entity_name'),
    ]

    operations = [
    ]



# Generated by Django 2.1 on 2019-07-22 12:32

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('djconnectwise', '0103_activitystatus_activitytype'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='activitystatus',
            options={'ordering': ('name',), 'verbose_name_plural': 'activity statuses'},
        ),
    ]



# Generated by Django 3.1 on 2020-09-23 09:52

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('djconnectwise', '0124_auto_20200917_1922'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='agreement',
            options={'get_latest_by': 'modified'},
        ),
        migrations.AlterModelOptions(
            name='projectteammember',
            options={'get_latest_by': 'modified'},
        ),
        migrations.AlterModelOptions(
            name='workrole',
            options={'get_latest_by': 'modified'},
        ),
        migrations.AlterModelOptions(
            name='worktype',
            options={'get_latest_by': 'modified'},
        ),
    ]



import logging
import datetime
import math


from dateutil.parser import parse
from copy import deepcopy
from decimal import Decimal

from django.core.files.storage import default_storage
from django.core.exceptions import ObjectDoesNotExist
from django.db import transaction, IntegrityError
from django.utils import timezone
from django.db.models import Q

from djconnectwise import api
from djconnectwise import models
from djconnectwise.utils import get_hash, get_filename_extension, \
    generate_thumbnail, generate_filename, remove_thumbnail
from djconnectwise.utils import DjconnectwiseSettings

from botocore.exceptions import NoCredentialsError

DEFAULT_AVATAR_EXTENSION = 'jpg'
MAX_URL_LENGTH = 2000
MIN_URL_LENGTH = 1980

CREATED = 1
UPDATED = 2
SKIPPED = 3

logger = logging.getLogger(__name__)


class InvalidObjectException(Exception):
    """
    If for any reason an object can't be created (for example, it references
    an unknown foreign object, or is missing a required field), raise this
    so that the synchronizer can catch it and continue with other records.
    """
    pass


def log_sync_job(f):
    def wrapper(*args, **kwargs):
        sync_instance = args[0]
        created_count = updated_count = deleted_count = skipped_count = 0
        sync_job = models.SyncJob()
        sync_job.start_time = timezone.now()
        sync_job.entity_name = sync_instance.model_class.__bases__[0].__name__
        sync_job.synchronizer_class = \
            sync_instance.__class__.__name__

        if sync_instance.full:
            sync_job.sync_type = 'full'
        else:
            sync_job.sync_type = 'partial'

        sync_job.save()

        try:
            created_count, updated_count, skipped_count, deleted_count = \
                f(*args, **kwargs)
            sync_job.success = True
        except Exception as e:
            sync_job.message = str(e.args[0])
            sync_job.success = False
            raise
        finally:
            sync_job.end_time = timezone.now()
            sync_job.added = created_count
            sync_job.updated = updated_count
            sync_job.skipped = skipped_count
            sync_job.deleted = deleted_count
            sync_job.save()

        return created_count, updated_count, skipped_count, deleted_count
    return wrapper


class SyncResults:
    """Track results of a sync job."""
    def __init__(self):
        self.created_count = 0
        self.updated_count = 0
        self.skipped_count = 0
        self.deleted_count = 0
        self.synced_ids = set()


class Synchronizer:
    lookup_key = 'id'

    def __init__(self, full=False, *args, **kwargs):
        self.api_conditions = []
        self.client = self.client_class()
        request_settings = DjconnectwiseSettings().get_settings()
        self.batch_size = request_settings['batch_size']
        self.full = full

        self.pre_delete_callback = kwargs.pop('pre_delete_callback', None)
        self.pre_delete_args = kwargs.pop('pre_delete_args', None)
        self.post_delete_callback = kwargs.pop('post_delete_callback', None)

    def set_relations(self, instance, json_data):
        for json_field, value in self.related_meta.items():
            model_class, field_name = value
            self._assign_relation(
                instance,
                json_data,
                json_field,
                model_class,
                field_name
            )

    @staticmethod
    def _assign_null_relation(instance, model_field):
        """
        Set the FK to null, but handle issues like the FK being non-null.

        This can happen because ConnectWise gives us records that point to
        non-existent records- such as activities whose assignTo fields point
        to deleted members.
        """
        try:
            setattr(instance, model_field, None)
        except ValueError:
            # The model_field may have been non-null.
            raise InvalidObjectException(
                "Unable to set field {} on {} to null, as it's required.".
                format(model_field, instance)
            )

    def _assign_relation(self, instance, json_data,
                         json_field, model_class, model_field):
        """
        Look up the given foreign relation, and set it to the given
        field on the instance.
        """
        relation_json = json_data.get(json_field)
        if relation_json is None:
            self._assign_null_relation(instance, model_field)
            return

        uid = relation_json['id']

        try:
            related_instance = model_class.objects.get(pk=uid)
            setattr(instance, model_field, related_instance)
        except model_class.DoesNotExist:
            logger.warning(
                'Failed to find {} {} for {} {}.'.format(
                    json_field,
                    uid,
                    type(instance),
                    instance.id
                )
            )
            self._assign_null_relation(instance, model_field)

    def _instance_ids(self, filter_params=None):
        if not filter_params:
            ids = self.model_class.objects.all().order_by(self.lookup_key)\
                .values_list(self.lookup_key, flat=True)
        else:
            ids = self.model_class.objects.filter(filter_params)\
                .order_by(self.lookup_key)\
                .values_list(self.lookup_key, flat=True)
        return set(ids)

    def get(self, results, conditions=None):
        return self.fetch_records(results, conditions)

    def fetch_records(self, results, conditions=None):
        """
        For all pages of results, save each page of results to the DB.

        If conditions is supplied in the call, then use only those conditions
        while fetching pages of records. If it's omitted, then use
        self.api_conditions.
        """
        page = 1
        while True:
            logger.info(
                'Fetching {} records, batch {}'.format(
                    self.model_class.__bases__[0].__name__, page)
            )
            page_conditions = conditions or self.api_conditions
            page_records = self.get_page(
                page=page, page_size=self.batch_size,
                conditions=page_conditions,
            )
            self.persist_page(page_records, results)
            page += 1
            if len(page_records) < self.batch_size:
                # This page wasn't full, so there's no more records after
                # this page.
                break
        return results

    def persist_page(self, records, results):
        """Persist one page of records to DB."""
        for record in records:
            try:
                with transaction.atomic():
                    instance, result = self.update_or_create_instance(record)
                if result == CREATED:
                    results.created_count += 1
                elif result == UPDATED:
                    results.updated_count += 1
                else:
                    results.skipped_count += 1
            except InvalidObjectException as e:
                logger.warning('{}'.format(e))

            results.synced_ids.add(record['id'])

        return results

    def get_page(self, *args, **kwargs):
        raise NotImplementedError

    def get_single(self, *args, **kwargs):
        raise NotImplementedError

    def _assign_field_data(self, instance, api_instance):
        raise NotImplementedError

    def fetch_sync_by_id(self, instance_id, sync_config={}):
        api_instance = self.get_single(instance_id)
        instance, created = self.update_or_create_instance(api_instance)
        return instance

    def fetch_delete_by_id(self, instance_id, pre_delete_callback=None,
                           pre_delete_args=None,
                           post_delete_callback=None):
        try:
            self.get_single(instance_id)
            logger.warning(
                'ConnectWise API returned {} {} even though it was expected '
                'to be deleted.'.format(
                    self.model_class.__bases__[0].__name__, instance_id)
            )

        except api.ConnectWiseRecordNotFoundError:
            # This is what we expect to happen. Since it's gone in CW, we
            # are safe to delete it from here.
            pre_delete_result = None
            try:
                if pre_delete_callback:
                    pre_delete_result = pre_delete_callback(*pre_delete_args)
                self.model_class.objects.filter(pk=instance_id).delete()
            finally:
                if post_delete_callback:
                    post_delete_callback(pre_delete_result)
            logger.info(
                'Deleted {} {} (if it existed).'.format(
                    self.model_class.__bases__[0].__name__,
                    instance_id
                )
            )

    def update_or_create_instance(self, api_instance):
        """
        Creates and returns an instance if it does not already exist.
        """
        result = None
        api_instance = self.remove_null_characters(api_instance)
        try:
            instance_pk = api_instance[self.lookup_key]
            instance = self.model_class.objects.get(pk=instance_pk)
        except self.model_class.DoesNotExist:
            instance = self.model_class()
            result = CREATED

        try:
            self._assign_field_data(instance, api_instance)

            # This will return the created instance, the updated instance, or
            # if the instance is skipped an unsaved copy of the instance.
            if result == CREATED:
                if self.model_class is models.TicketTracker:
                    instance.save(force_insert=True)
                else:
                    instance.save()
            elif instance.tracker.changed():
                instance.save()
                result = UPDATED
            else:
                result = SKIPPED
        except IntegrityError as e:
            # This can happen when multiple threads are creating the
            # same ticket at once. See issue description for #991
            # for the full details.
            msg = "IntegrityError while attempting to create {}." \
                  " Error: {}".format(self.model_class, e)
            logger.error(msg)
            raise InvalidObjectException(msg)

        if result == CREATED:
            result_log = 'Created'
        elif result == UPDATED:
            result_log = 'Updated'
        else:
            result_log = 'Skipped'

        logger.info('{}: {} {}'.format(
            result_log,
            self.model_class.__bases__[0].__name__,
            instance
        ))

        return instance, result

    def prune_stale_records(self, initial_ids, synced_ids):
        """
        Delete records that existed when sync started but were
        not seen as we iterated through all records from REST API.
        """
        stale_ids = initial_ids - synced_ids
        deleted_count = 0
        if stale_ids:
            delete_qset = self.get_delete_qset(stale_ids)
            deleted_count = delete_qset.count()

            pre_delete_result = None
            if self.pre_delete_callback:
                pre_delete_result = self.pre_delete_callback(
                    *self.pre_delete_args
                )
            logger.info(
                'Removing {} stale records for model: {}'.format(
                    len(stale_ids), self.model_class.__bases__[0].__name__,
                )
            )
            delete_qset.delete()
            if self.post_delete_callback:
                self.post_delete_callback(pre_delete_result)

        return deleted_count

    def get_delete_qset(self, stale_ids):
        return self.model_class.objects.filter(pk__in=stale_ids)

    def get_sync_job_qset(self):
        return models.SyncJob.objects.filter(
            entity_name=self.model_class.__bases__[0].__name__
        )

    @log_sync_job
    def sync(self):
        sync_job_qset = self.get_sync_job_qset()

        # Since the job is created before it begins, make sure to exclude
        # itself, and at least one other sync job exists.
        if sync_job_qset.count() > 1 and not self.full:
            last_sync_job_time = sync_job_qset.exclude(
                id=sync_job_qset.last().id).last().start_time.isoformat()
            self.api_conditions.append(
                "lastUpdated>[{0}]".format(last_sync_job_time)
            )
        results = SyncResults()
        results = self.get(results, )

        if self.full:
            # Set of IDs of all records prior to sync,
            # to find stale records for deletion.
            initial_ids = self._instance_ids()

            results.deleted_count = self.prune_stale_records(
                initial_ids, results.synced_ids
            )

        return results.created_count, results.updated_count, \
            results.skipped_count, results.deleted_count

    def callback_sync(self, filter_params):

        results = SyncResults()

        # Set of IDs of all records related to the parent object
        # to sync, to find stale records for deletion, need to be careful
        # to get the correct filter on the objects you want, or you will
        # be deleting records you didn't intend to, and they wont be restored
        # until their next scheduled sync
        initial_ids = self._instance_ids(filter_params=filter_params)
        results = self.get(results, )

        # This should always be a full sync (unless something changes in
        # the future and it doesn't need to delete anything)
        results.deleted_count = self.prune_stale_records(
            initial_ids, results.synced_ids
        )

        return results.created_count, results.updated_count, \
            results.skipped_count, results.deleted_count

    def sync_children(self, *args):
        for synchronizer, filter_params in args:
            created_count, updated_count, skipped_count, deleted_count \
                = synchronizer.callback_sync(filter_params)
            msg = '{} Child Sync - Created: {},'\
                ' Updated: {}, Skipped: {},Deleted: {}'.format(
                    synchronizer.model_class.__bases__[0].__name__,
                    created_count,
                    updated_count,
                    skipped_count,
                    deleted_count
                )
            logger.info(msg)

    def remove_null_characters(self, json_data):
        for value in json_data:
            if isinstance(json_data.get(value), str):
                json_data[value] = json_data[value].replace('\x00', '')

        return json_data


class BatchConditionMixin:
    """
    Yo I heard you like pages, so I put pages in your pages!

    This mixin has methods for Synchronizers that have to page their pagers.
    For example, syncing the tickets in a set of statuses- there can be
    too many statuses to fit all their IDs in one URL, given the max URL
    size of approximately 2000 characters. So we split the statuses into
    groups that get the URL close to 2000 characters, and get those results
    in pages. And then get the next set of statuses in pages, and so on.
    """
    def get_batch_condition(self, conditions):
        raise NotImplementedError

    def get(self, results, conditions=None):
        """Buffer and return all pages of results."""
        unfetched_conditions = deepcopy(self.batch_condition_list)
        while unfetched_conditions:
            # While there are still items left in the list there are still
            # batches left to be fetched.
            optimal_size = self.get_optimal_size(unfetched_conditions)
            batch_conditions = unfetched_conditions[:optimal_size]
            del unfetched_conditions[:optimal_size]
            batch_condition = self.get_batch_condition(batch_conditions)
            batch_conditions = deepcopy(self.api_conditions)
            batch_conditions.append(batch_condition)
            results = super().get(results, conditions=batch_conditions)
        return results

    def get_optimal_size(self, condition_list):
        if not condition_list:
            # Return none if empty list
            return None
        size = len(condition_list)
        if self.url_length(condition_list, size) < MAX_URL_LENGTH:
            # If we can fit all of the statuses in the first batch, return
            return size

        max_size = size
        min_size = 1
        while True:
            url_len = self.url_length(condition_list, size)
            if url_len <= MAX_URL_LENGTH and url_len > MIN_URL_LENGTH:
                break
            elif url_len > MAX_URL_LENGTH:
                max_size = size
                size = math.floor((max_size+min_size)/2)
            else:
                min_size = size
                size = math.floor((max_size+min_size)/2)
            if min_size == 1 and max_size == 1:
                # The URL cannot be made short enough. This ought never to
                # happen in production.
                break
        return size

    @staticmethod
    def url_length(condition_list, size):
        # We add the approximate amount of characters for the URL that we
        # know will be there every time (200) plus a little more to ensure we
        # don't undercut it (300), and add size, because for the amount of
        # conditions, there will be just as many commas separating them
        return sum(len(str(i)) for i in condition_list[:size]) + 300 + size


class CallbackPartialSyncMixin:
    """
    Run a partial sync on callbacks for related synchronizers. If a ticket
    has many notes or time entries they could all be fetched
    when a callback runs. This could mean a lot of time and resources spent
    syncing old notes or time entries. We will continue to return
    deleted_count even though it will always be zero in this case.
    """
    def callback_sync(self, filter_params):
        sync_job_qset = self.get_sync_job_qset()

        if sync_job_qset.exists():
            last_sync_job_time = sync_job_qset.last().start_time.isoformat()
            self.api_conditions.append(
                "lastUpdated>[{0}]".format(last_sync_job_time)
            )

        results = SyncResults()
        results = self.get(results, )

        return results.created_count, results.updated_count, \
            results.skipped_count, results.deleted_count


class ServiceNoteSynchronizer(CallbackPartialSyncMixin, Synchronizer):
    client_class = api.ServiceAPIClient
    model_class = models.ServiceNoteTracker

    related_meta = {
        'member': (models.Member, 'member')
    }

    def _assign_field_data(self, instance, json_data):
        instance.id = json_data['id']
        instance.text = json_data.get('text')
        instance.detail_description_flag = json_data.get(
            'detailDescriptionFlag')
        instance.internal_analysis_flag = json_data.get(
            'internalAnalysisFlag')
        instance.resolution_flag = json_data.get(
            'resolutionFlag')
        instance.created_by = json_data.get('createdBy')
        instance.internal_flag = json_data.get('internalFlag')
        instance.external_flag = json_data.get('externalFlag')

        date_created = json_data.get('dateCreated')
        if date_created:
            instance.date_created = parse(date_created)

        ticket_class = models.Ticket

        try:
            ticket_id = json_data['ticketId']
            related_ticket = ticket_class.objects.get(pk=ticket_id)
            setattr(instance, 'ticket', related_ticket)
        except KeyError:
            raise InvalidObjectException(
                'Service note {} has no ticketId key to find its target'
                '- skipping.'.format(instance.id)
            )
        except ObjectDoesNotExist as e:
            raise InvalidObjectException(
                'Service note {} has a ticketId that does not exist.'
                ' ObjectDoesNotExist Exception: {}'.format(instance.id, e)
            )

        self.set_relations(instance, json_data)

    def client_call(self, ticket_id, *args, **kwargs):
        return self.client.get_notes(ticket_id, *args, **kwargs)

    def get_page(self, *args, **kwargs):
        records = []
        ticket_qs = models.Ticket.objects.all().order_by(self.lookup_key)

        # We are using the conditions here to specify getting a single
        # tickets notes. If conditions are supplied, like when a ticket
        # callback is fired, then replace the conditions with the
        # last updated sync time instead of syncing all notes all the time.
        conditions = kwargs.get('conditions')
        if conditions:
            try:
                ticket_id = int(conditions[0])
                kwargs['conditions'] = \
                    [conditions[1]] if len(conditions) > 1 else []

                records += self.client_call(ticket_id, *args, **kwargs)
                return records
            except ValueError:
                # Do nothing
                pass
        for ticket_id in ticket_qs.values_list('id', flat=True):
            records += self.client_call(ticket_id, *args, **kwargs)

        return records

    def create_new_note(self, target, **kwargs):
        """
        Send POST request to ConnectWise to create a new note and then
        create it in the local database from the response
        """

        target_data = {}

        if isinstance(target, models.Ticket):
            target_data['id'] = target.id
            target_data['type'] = target.record_type
        else:
            raise ValueError(
                "Invalid target type for note creation: {}.".format(
                    str(target.__class__))
            )
        service_client = api.ServiceAPIClient(
            api_public_key=kwargs.get('api_public_key'),
            api_private_key=kwargs.get('api_private_key')
        )
        instance = service_client.post_note(target_data, **kwargs)

        return self.update_or_create_instance(instance)


class OpportunityNoteSynchronizer(Synchronizer):
    client_class = api.SalesAPIClient
    model_class = models.OpportunityNoteTracker

    def _assign_field_data(self, instance, json_data):
        instance.id = json_data['id']
        instance.text = json_data.get('text')

        opp_class = models.Opportunity

        date_created = json_data.get('_info').get('lastUpdated')

        if date_created:
            instance.date_created = parse(date_created)

        try:
            opportunity_id = json_data.get('opportunityId')
            related_opportunity = opp_class.objects.get(pk=opportunity_id)
            setattr(instance, 'opportunity', related_opportunity)
        except ObjectDoesNotExist as e:
            raise InvalidObjectException(
                'Opportunity note {} has a opportunityId that does not exist.'
                ' ObjectDoesNotExist Exception: {}'.format(instance.id, e)
            )

    def client_call(self, opportunity_id, *args, **kwargs):
        return self.client.get_notes(opportunity_id, *args, **kwargs)

    def get_page(self, *args, **kwargs):
        records = []
        opportunity_qs = models.Opportunity.objects.all()\
            .order_by(self.lookup_key)

        for opportunity_id in opportunity_qs.values_list('id', flat=True):
            records += self.client_call(opportunity_id, *args, **kwargs)

        return records


class BoardSynchronizer(Synchronizer):
    client_class = api.ServiceAPIClient
    model_class = models.ConnectWiseBoardTracker

    related_meta = {
        'workRole': (models.WorkRole, 'work_role'),
        'workType': (models.WorkType, 'work_type')
    }

    def _assign_field_data(self, instance, json_data):
        instance.id = json_data['id']
        instance.name = json_data['name']
        if json_data['billTime'] == 'NoDefault':
            instance.bill_time = None
        else:
            instance.bill_time = json_data['billTime']

        if 'inactiveFlag' in json_data:
            # This is the new CW way
            instance.inactive = json_data.get('inactiveFlag')
        else:
            # This is old, but keep for backwards-compatibility
            instance.inactive = json_data.get('inactive')

        instance.project_flag = json_data.get('projectFlag', False)

        self.set_relations(instance, json_data)
        return instance

    def get_page(self, *args, **kwargs):
        return self.client.get_boards(*args, **kwargs)


class BoardChildSynchronizer(Synchronizer):

    def _assign_field_data(self, instance, json_data):
        instance.id = json_data['id']
        instance.name = json_data['name']
        try:
            board_id = json_data['board']['id']
        except KeyError:
            # Must be 2017.5 or earlier
            board_id = json_data['boardId']
        instance.board = models.ConnectWiseBoard.objects.get(id=board_id)
        return instance

    def client_call(self, board_id):
        raise NotImplementedError


class BoardStatusSynchronizer(BoardChildSynchronizer):
    client_class = api.ServiceAPIClient
    model_class = models.BoardStatusTracker

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.api_conditions = ['inactive=False']

    def _assign_field_data(self, instance, json_data):
        instance = super(BoardStatusSynchronizer, self)._assign_field_data(
            instance, json_data)

        instance.sort_order = json_data.get('sortOrder')
        instance.display_on_board = json_data.get('displayOnBoard')
        instance.inactive = json_data.get('inactive')
        instance.closed_status = json_data.get('closedStatus')
        instance.escalation_status = json_data.get('escalationStatus')
        instance.time_entry_not_allowed = json_data.get('timeEntryNotAllowed')

        return instance

    def client_call(self, board_id, *args, **kwargs):
        kwargs['conditions'] = self.api_conditions
        return self.client.get_statuses(board_id, *args, **kwargs)

    def get_page(self, *args, **kwargs):
        records = []
        board_qs = models.ConnectWiseBoard.objects.all()\
            .order_by(self.lookup_key)

        for board_id in board_qs.values_list('id', flat=True):
            records += self.client_call(board_id, *args, **kwargs)

        return records


class BoardFilterMixin:

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        request_settings = DjconnectwiseSettings().get_settings()
        board_names = request_settings.get('board_status_filter')
        self.boards = [board.strip() for board in board_names.split(',')] \
            if board_names else None

    def get_page(self, *args, **kwargs):
        records = []
        if self.boards:
            board_qs = models.ConnectWiseBoard.available_objects.filter(
                name__in=self.boards).order_by(self.lookup_key)
        else:
            board_qs = models.ConnectWiseBoard.available_objects.all().\
                order_by(self.lookup_key)

        for board_id in board_qs.values_list('id', flat=True):
            records += self.client_call(board_id, *args, **kwargs)

        return records


class TeamSynchronizer(BoardFilterMixin, BoardChildSynchronizer):
    client_class = api.ServiceAPIClient
    model_class = models.TeamTracker

    def _assign_field_data(self, instance, json_data):
        instance = super(TeamSynchronizer, self)._assign_field_data(
            instance, json_data)

        members = []
        if json_data.get('members'):
            members = list(models.Member.objects.filter(
                id__in=json_data['members']))

        instance.save()

        instance.members.clear()
        instance.members.add(*members)
        return instance

    def client_call(self, board_id, *args, **kwargs):
        return self.client.get_teams(board_id, *args, **kwargs)


class CompanySynchronizer(Synchronizer):
    client_class = api.CompanyAPIClient
    model_class = models.CompanyTracker

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.api_conditions = ['deletedFlag=False']

    def _assign_field_data(self, company, company_json):
        """
        Assigns field data from an company_json instance
        to a local Company model instance
        """
        company.id = company_json['id']
        company.name = company_json['name']
        company.identifier = company_json['identifier']

        # Fields below aren't included when the company is created as a
        # side-effect of creating/updating a ticket or other type of object,
        # so use .get().
        company.phone_number = company_json.get('phoneNumber')
        company.fax_number = company_json.get('faxNumber')
        company.address_line1 = company_json.get('addressLine1')
        company.address_line2 = company_json.get('addressLine2')
        company.city = company_json.get('city')
        company.state_identifier = company_json.get('state')
        company.zip = company_json.get('zip')
        company.deleted_flag = company_json.get('deletedFlag', False)

        status_json = company_json.get('status')
        if status_json:
            try:
                status = models.CompanyStatus.objects.get(pk=status_json['id'])
                company.status = status
            except models.CompanyStatus.DoesNotExist:
                logger.warning(
                    'Failed to find CompanyStatus: {}'.format(
                        status_json['id']
                    ))

        calendar_id = company_json.get('calendarId')
        if calendar_id:
            try:
                company.calendar = models.Calendar.objects.get(id=calendar_id)
            except models.Calendar.DoesNotExist:
                logger.warning(
                    'Failed to find Calendar: {}'.format(
                        calendar_id
                    ))
        else:
            company.calendar = calendar_id

        types_list = company_json.get('typeIds', [])
        for type_id in types_list:
            try:
                company_type = models.CompanyType.objects.get(
                    pk=type_id)
                company.company_types.add(company_type)
            except models.CompanyType.DoesNotExist:
                logger.warning(
                    'Failed to find CompanyType: {}'.format(
                        type_id
                    )
                )

        territory_id = company_json.get('territoryId')
        if territory_id:
            try:
                company.territory = models.Territory.objects.get(
                    pk=territory_id
                )
            except models.Territory.DoesNotExist:
                logger.warning(
                    'Failed to find Territory: {}'.format(
                        territory_id
                    )
                )
        else:
            logger.warning(
                'No Territory ID recieved in request for Company: {}'.format(
                    company.id
                )
            )
        return company

    def get_page(self, *args, **kwargs):
        return self.client.get_companies(*args, **kwargs)

    def get_single(self, company_id):
        return self.client.by_id(company_id)

    def fetch_delete_by_id(self, company_id, pre_delete_callback=None,
                           pre_delete_args=None,
                           post_delete_callback=None):
        # Companies are deleted by setting deleted_flag = True, so
        # just treat this as a normal sync.
        pre_delete_result = None
        try:
            if pre_delete_callback:
                pre_delete_result = pre_delete_callback(*pre_delete_args)
            self.fetch_sync_by_id(company_id)
        finally:
            if post_delete_callback:
                post_delete_callback(pre_delete_result)


class CompanyStatusSynchronizer(Synchronizer):
    client_class = api.CompanyAPIClient
    model_class = models.CompanyStatusTracker

    def _assign_field_data(self, instance, json_data):
        instance.id = json_data['id']
        instance.name = json_data['name']
        instance.default_flag = json_data.get('defaultFlag')
        instance.inactive_flag = json_data.get('inactiveFlag')
        instance.notify_flag = json_data.get('notifyFlag')
        instance.dissalow_saving_flag = json_data.get('disallowSavingFlag')
        instance.notification_message = json_data.get('notificationMessage')
        instance.custom_note_flag = json_data.get('customNoteFlag')
        instance.cancel_open_tracks_flag = json_data.get(
            'cancelOpenTracksFlag'
        )

        if json_data.get('track'):
            instance.track_id = json_data['track']['id']

        return instance

    def get_page(self, *args, **kwargs):
        return self.client.get_company_statuses(*args, **kwargs)


class CompanyTypeSynchronizer(Synchronizer):
    client_class = api.CompanyAPIClient
    model_class = models.CompanyTypeTracker

    def _assign_field_data(self, instance, json_data):
        instance.id = json_data['id']
        instance.name = json_data.get('name')
        instance.vendor_flag = json_data['vendorFlag']

    def get_page(self, *args, **kwargs):
        return self.client.get_company_types(*args, **kwargs)


class MyCompanyOtherSynchronizer(Synchronizer):
    client_class = api.SystemAPIClient
    model_class = models.MyCompanyOtherTracker

    related_meta = {
        'defaultCalendar': (models.Calendar, 'default_calendar'),
    }

    def _assign_field_data(self, instance, json_data):
        instance.id = json_data['id']
        self.set_relations(instance, json_data)
        return instance

    def get_page(self, *args, **kwargs):
        return self.client.get_mycompanyother(*args, **kwargs)


class ActivityStatusSynchronizer(Synchronizer):
    client_class = api.SalesAPIClient
    model_class = models.ActivityStatusTracker

    def _assign_field_data(self, instance, json_data):
        instance.id = json_data['id']
        instance.name = json_data['name']
        instance.default_flag = json_data.get('defaultFlag', False)
        instance.inactive_flag = json_data.get('inactiveFlag', False)
        instance.spawn_followup_flag = \
            json_data.get('spawnFollowupFlag', False)
        instance.closed_flag = json_data.get('closedFlag', False)
        return instance

    def get_page(self, *args, **kwargs):
        return self.client.get_activity_statuses(*args, **kwargs)


class ActivityTypeSynchronizer(Synchronizer):
    client_class = api.SalesAPIClient
    model_class = models.ActivityTypeTracker

    def _assign_field_data(self, instance, json_data):
        instance.id = json_data['id']
        instance.name = json_data['name']
        instance.points = json_data['points']
        instance.default_flag = json_data.get('defaultFlag', False)
        instance.inactive_flag = json_data.get('inactiveFlag', False)
        instance.email_flag = json_data.get('emailFlag', False)
        instance.memo_flag = json_data.get('memoFlag', False)
        instance.history_flag = json_data.get('historyFlag', False)

        return instance

    def get_page(self, *args, **kwargs):
        return self.client.get_activity_types(*args, **kwargs)


class ActivitySynchronizer(Synchronizer):
    client_class = api.SalesAPIClient
    model_class = models.ActivityTracker

    related_meta = {
        'opportunity': (models.Opportunity, 'opportunity'),
        'ticket': (models.Ticket, 'ticket'),
        'assignTo': (models.Member, 'assign_to'),
        'status': (models.ActivityStatus, 'status'),
        'type': (models.ActivityType, 'type'),
        'company': (models.Company, 'company'),
        'agreement': (models.Agreement, 'agreement'),
    }

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # Only sync activities in non-closed statuses. There shouldn't be
        # too many activity statuses so we don't need to page this like we
        # do with tickets.
        self.api_conditions = ['status/id in ({})'.format(
            ','.join(
                [
                    str(i.id) for
                    i in models.ActivityStatus.objects.filter(
                        closed_flag=False
                    )
                ]
            )
        )]

    def _assign_field_data(self, instance, json_data):
        instance.id = json_data['id']
        instance.name = json_data['name']
        instance.notes = json_data.get('notes')

        # handle dates.  Assume UTC timezone when not defined
        # (according to ConnectWise FAQ: "What DateTimes are supported?")
        date_start = json_data.get('dateStart')
        if date_start:
            instance.date_start = parse(date_start, default=parse('00:00Z'))

        date_end = json_data.get('dateEnd')
        if date_end:
            instance.date_end = parse(date_end, default=parse('00:00Z'))

        # Creating activities in the ConnectWise UI and API requires that
        # the 'assignTo' field is set. But we've seen cases where
        # 'assignTo' is null. So skip the activity if it's 'assignTo' is null.
        assign_to_id = json_data.get('assignTo')
        if not assign_to_id:
            raise InvalidObjectException(
                'Activity {} has a null assignTo field - '
                'skipping.'.format(instance.id)
            )

        instance.udf = {str(item['id']): item
                        for item in json_data.get('customFields', list())}

        self.set_relations(instance, json_data)
        return instance

    def get_page(self, *args, **kwargs):
        return self.client.get_activities(*args, **kwargs)

    def get_single(self, activity_id):
        return self.client.get_single_activity(activity_id)


class SalesProbabilitySynchronizer(Synchronizer):
    client_class = api.SalesAPIClient
    model_class = models.SalesProbabilityTracker

    def _assign_field_data(self, instance, json_data):
        instance.id = json_data['id']
        instance.probability = json_data['probability']

    def get_page(self, *args, **kwargs):
        return self.client.get_probabilities(*args, **kwargs)


class ScheduleEntriesSynchronizer(BatchConditionMixin, Synchronizer):
    client_class = api.ScheduleAPIClient
    model_class = models.ScheduleEntryTracker
    batch_condition_list = []

    related_meta = {
        'where': (models.Location, 'where'),
        'status': (models.ScheduleStatus, 'status'),
        'type': (models.ScheduleType, 'schedule_type'),
        'member': (models.Member, 'member')
    }

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.settings = DjconnectwiseSettings().get_settings()
        self.no_batch = kwargs.get('no_batch')

        self.api_conditions = [
            "(type/identifier='S' or type/identifier='O')"
        ]

        if not self.no_batch:
            self.api_conditions.append("doneFlag=false")

            # Only get schedule entries for tickets or opportunities that we
            # already have in the DB.
            ticket_ids = set(
                models.Ticket.objects.order_by(
                    self.lookup_key).values_list('id', flat=True)
            )
            opportunity_ids = set(
                models.Opportunity.objects.order_by(
                    self.lookup_key).values_list('id', flat=True)
            )
            self.batch_condition_list = list(ticket_ids | opportunity_ids)

    def get(self, results, conditions=None):

        if self.no_batch:
            self.fetch_records(results, conditions)
        else:
            super().get(results, conditions)

        return results

    def get_optimal_size(self, condition_list):
        object_id_size = self.settings['schedule_entry_conditions_size']

        return object_id_size if object_id_size \
            else super().get_optimal_size(condition_list)

    def get_batch_condition(self, conditions):
        return 'objectId in ({})'.format(
            ','.join([str(i) for i in conditions])
        )

    def _assign_field_data(self, instance, json_data):
        instance.id = json_data['id']
        instance.name = json_data.get('name')
        instance.done_flag = json_data['doneFlag']

        # Handle member- there are cases where no member field is provided,
        # which we consider invalid.
        if 'member' not in json_data:
            raise InvalidObjectException(
                'Schedule entry {} has no member field- skipping.'
                .format(instance.id)
            )

        try:
            member_id = json_data['member']['id']
            models.Member.objects.get(pk=member_id)
        except ObjectDoesNotExist:
            raise InvalidObjectException(
                'Schedule entry {} can not find member: {}'
                '- skipping.'.format(instance.id, member_id)
            )

        # handle dates
        # Connectwise handles schedule entries with a date and no time by
        # leaving it as midnight UTC. This has the unfortunate side effect of
        # looking like a completely valid time in other time zones. Ex, In
        # UTC-8 it looks like 4pm - 4pm. Alter these midnight datetimes to
        # appear as midnight in the local time, instead of midnight UTC.
        date_start = json_data.get('dateStart')
        if date_start:
            instance.date_start = parse(date_start)

            date_end = json_data.get('dateEnd')

            if instance.date_start.time() == datetime.time(0, 0) \
                    and json_data.get('hours') == 0.0:
                instance.date_start = parse(date_start) - \
                    timezone.localtime().utcoffset()
                if date_end:
                    instance.date_end = parse(date_end) - \
                        timezone.localtime().utcoffset()
            elif date_end:
                instance.date_end = parse(date_end)

        self.set_relations(instance, json_data)

        ticket_class = models.Ticket
        activity_class = models.Activity
        try:
            uid = json_data['objectId']
        except KeyError:
            raise InvalidObjectException(
                'Schedule entry {} has no objectId key to find its target'
                '- skipping.'.format(instance.id)
            )

        if json_data['type']['identifier'] == "S":
            try:
                related_ticket = ticket_class.objects.get(pk=uid)
                if json_data['doneFlag']:
                    setattr(instance, 'ticket_object', None)
                else:
                    setattr(instance, 'ticket_object', related_ticket)
            except ObjectDoesNotExist as e:
                logger.warning(
                    'Ticket not found for {}.'.format(instance.id) +
                    ' ObjectDoesNotExist Exception: {}.'.format(e)
                )
        elif json_data['type']['identifier'] == "C":
            try:
                related_activity = activity_class.objects.get(pk=uid)
                setattr(instance, 'activity_object', related_activity)
            except ObjectDoesNotExist as e:
                logger.warning(
                    'Activity not found for {}.'.format(instance.id) +
                    ' ObjectDoesNotExist Exception: {}.'.format(e)
                )
        else:
            raise InvalidObjectException(
                'Invalid ScheduleEntry type for schedule entry {}- skipping.'
                .format(instance.id)
            )

        return instance

    def get_page(self, *args, **kwargs):
        return self.client.get_schedule_entries(*args, **kwargs)

    def get_single(self, entry_id):
        return self.client.get_schedule_entry(entry_id)

    def create_new_entry(self, target, **kwargs):
        """
        Send POST request to ConnectWise to create a new entry and then
        create it in the local database from the response

        """
        schedule_client = api.ScheduleAPIClient(
            api_public_key=kwargs.get('api_public_key'),
            api_private_key=kwargs.get('api_private_key')
        )
        # Type defines if it is a service ticket, opportunity, or activity.
        # (There are more but we probably wont be using them).
        # In the context of schedule entry types, Service Tickets and
        # Project tickets are the same.
        schedule_type = models.ScheduleType.objects.get(
            identifier=target.SCHEDULE_ENTRY_TYPE)

        instance = schedule_client.post_schedule_entry(
            target, schedule_type, **kwargs)
        return self.update_or_create_instance(instance)

    def update_entry(self, **kwargs):
        """
        Send PATCH request to ConnectWise to update an entry and then
        update it in the local database from the response
        """
        schedule_client = api.ScheduleAPIClient()
        instance = schedule_client.patch_schedule_entry(**kwargs)
        return self.update_or_create_instance(instance)


class ScheduleStatusSynchronizer(Synchronizer):
    client_class = api.ScheduleAPIClient
    model_class = models.ScheduleStatusTracker

    def _assign_field_data(self, instance, json_data):
        instance.id = json_data['id']
        instance.name = json_data['name']

        return instance

    def get_page(self, *args, **kwargs):
        return self.client.get_schedule_statuses(*args, **kwargs)


class ScheduleTypeSynchronizer(Synchronizer):
    client_class = api.ScheduleAPIClient
    model_class = models.ScheduleTypeTracker

    def _assign_field_data(self, instance, json_data):
        instance.id = json_data['id']
        instance.name = json_data['name']
        instance.identifier = json_data['identifier']

        return instance

    def get_page(self, *args, **kwargs):
        return self.client.get_schedule_types(*args, **kwargs)


class TerritorySynchronizer(Synchronizer):
    client_class = api.SystemAPIClient
    model_class = models.TerritoryTracker

    def _assign_field_data(self, instance, json_data):
        instance.id = json_data['id']
        instance.name = json_data['name']

        return instance

    def get_page(self, *args, **kwargs):
        return self.client.get_territories(*args, **kwargs)


class TimeEntrySynchronizer(BatchConditionMixin,
                            CallbackPartialSyncMixin, Synchronizer):
    client_class = api.TimeAPIClient
    model_class = models.TimeEntryTracker
    batch_condition_list = []

    related_meta = {
        'company': (models.Company, 'company'),
        'member': (models.Member, 'member')
    }

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.api_conditions = [
            "(chargeToType='ServiceTicket' OR chargeToType='ProjectTicket')"
        ]
        # Only get time entries for tickets that are already in the DB
        # Possibly Activities also in the future
        ticket_ids = set(
            models.Ticket.objects.order_by(
                self.lookup_key).values_list('id', flat=True)
        )
        self.batch_condition_list = list(ticket_ids)

    def get_batch_condition(self, conditions):
        return 'chargeToId in ({})'.format(
            ','.join([str(i) for i in conditions])
        )

    def _assign_field_data(self, instance, json_data):
        instance.id = json_data['id']
        instance.charge_to_type = json_data['chargeToType']
        instance.billable_option = json_data.get('billableOption')
        instance.notes = json_data.get('notes')
        instance.internal_notes = json_data.get('internalNotes')

        time_start = json_data.get('timeStart')
        if time_start:
            instance.time_start = parse(time_start, default=parse('00:00Z'))

        time_end = json_data.get('timeEnd')
        if time_end:
            instance.time_end = parse(time_end, default=parse('00:00Z'))

        # Since django's `to_python` method builds the record from the DB
        # by creating the decimal using a string we must also do this.
        # Normally it would not matter too much, but since we want to compare
        # them exactly it must be done this way. Otherwise due to floating
        # point errors the decimal we create with a float vs a string will
        # be different and fail comparison in the FieldTracker.
        hours_deduct = json_data.get('hoursDeduct')
        instance.hours_deduct = Decimal(str(hours_deduct)) \
            if hours_deduct is not None else None

        actual_hours = json_data.get('actualHours')
        instance.actual_hours = Decimal(str(actual_hours)) \
            if actual_hours is not None else None

        detail_description_flag = json_data.get('addToDetailDescriptionFlag')
        if detail_description_flag:
            instance.detail_description_flag = detail_description_flag

        internal_analysis_flag = json_data.get('addToInternalAnalysisFlag')
        if internal_analysis_flag:
            instance.internal_analysis_flag = internal_analysis_flag

        resolution_flag = json_data.get('addToResolutionFlag')
        if resolution_flag:
            instance.resolution_flag = resolution_flag

        self.set_relations(instance, json_data)

        # Similar to Schedule Entries, chargeToId is stored as an int in
        # ConnectWise, handled as special situation.
        # Not making a method to handle this in a similar way to Schedule
        # entries and even with the similar code as this may be VERY
        # different in the near future, because charge_to_id would be
        # converted to a GenericForeignKey and would be handled differently.
        ticket_class = models.Ticket
        try:
            charge_id = json_data['chargeToId']
        except KeyError:
            raise InvalidObjectException(
                'Time Entry {} has no chargeToId key to find its target'
                '- skipping.'.format(instance.id)
            )

        try:
            related_ticket = ticket_class.objects.get(pk=charge_id)
            setattr(instance, 'charge_to_id', related_ticket)
        except ObjectDoesNotExist as e:
            logger.warning(
                'Ticket not found for {}.'.format(instance.id) +
                ' ObjectDoesNotExist Exception: {}.'.format(e)
            )

        return instance

    def get_page(self, *args, **kwargs):
        return self.client.get_time_entries(*args, **kwargs)

    def create_new_entry(self, target, **kwargs):
        """
        Send POST request to ConnectWise to create a new entry and then
        create it in the local database from the response
        """
        target_data = {}
        if isinstance(target, models.Ticket):
            target_data['id'] = target.id
            record_type = target.record_type

            # ConnectWise doesn't allow "ProjectIssue" as a valid chargeToType.
            # See https://developer.connectwise.com/Products/Manage/REST#/TimeEntries # noqa
            # for valid chargeToTypes.
            if record_type == models.Ticket.PROJECT_ISSUE:
                record_type = models.Ticket.PROJECT_TICKET

            target_data['type'] = record_type
        else:
            raise InvalidObjectException(
                "Invalid target type for TimeEntry "
                "creation: " + str(target.__class__) + "."
            )

        time_client = api.TimeAPIClient(
            api_public_key=kwargs.get('api_public_key'),
            api_private_key=kwargs.get('api_private_key')
        )
        instance = time_client.post_time_entry(target_data, **kwargs)
        return self.update_or_create_instance(instance)


class LocationSynchronizer(Synchronizer):
    client_class = api.ServiceAPIClient
    model_class = models.LocationTracker

    def _assign_field_data(self, location, location_json):
        """
        Assigns field data from an company_json instance
        to a local Company model instance
        """
        location.id = location_json['id']
        location.name = location_json['name']
        location.where = location_json.get('where')
        return location

    def get_page(self, *args, **kwargs):
        return self.client.get_locations(*args, **kwargs)


class PrioritySynchronizer(Synchronizer):
    client_class = api.ServiceAPIClient
    model_class = models.TicketPriorityTracker

    def _assign_field_data(self, ticket_priority, api_priority):
        ticket_priority.id = api_priority['id']
        ticket_priority.name = api_priority['name']
        ticket_priority.color = api_priority.get('color')

        # work around due to api data inconsistencies
        sort_value = api_priority.get('sort') or api_priority.get('sortOrder')
        if sort_value:
            ticket_priority.sort = sort_value

        return ticket_priority

    def get_page(self, *args, **kwargs):
        return self.client.get_priorities(*args, **kwargs)


class ProjectStatusSynchronizer(Synchronizer):

    client_class = api.ProjectAPIClient
    model_class = models.ProjectStatusTracker

    def _assign_field_data(self, instance, json_data):
        instance.id = json_data['id']
        instance.name = json_data['name']
        instance.default_flag = json_data.get('defaultFlag')
        instance.inactive_flag = json_data.get('inactiveFlag')
        instance.closed_flag = json_data.get('closedFlag')
        return instance

    def get_page(self, *args, **kwargs):
        return self.client.get_project_statuses(*args, **kwargs)


class ProjectPhaseSynchronizer(Synchronizer):
    client_class = api.ProjectAPIClient
    model_class = models.ProjectPhaseTracker
    related_meta = {
        'board': (models.ConnectWiseBoard, 'board')
    }

    def _assign_field_data(self, instance, json_data):
        instance.id = json_data['id']
        instance.description = json_data['description']
        instance.notes = json_data.get('notes')
        instance.wbs_code = json_data.get('wbsCode')
        scheduled_hours = json_data.get('scheduledHours')
        budget_hours = json_data.get('budgetHours')
        actual_hours = json_data.get('actualHours')
        instance.budget_hours = Decimal(str(budget_hours)) \
            if budget_hours is not None else None
        instance.actual_hours = Decimal(str(actual_hours)) \
            if actual_hours is not None else None
        instance.scheduled_hours = Decimal(str(scheduled_hours)) \
            if scheduled_hours is not None else None

        if json_data['billTime'] == 'NoDefault':
            instance.bill_time = None
        else:
            instance.bill_time = json_data['billTime']

        scheduled_start = json_data.get('scheduledStart')
        scheduled_end = json_data.get('scheduledEnd')
        actual_start = json_data.get('actualStart')
        actual_end = json_data.get('actualEnd')

        if scheduled_start:
            instance.scheduled_start = parse(scheduled_start).date()

        if scheduled_end:
            instance.scheduled_end = parse(scheduled_end).date()

        if actual_start:
            instance.actual_start = parse(actual_start).date()

        if actual_end:
            instance.actual_end = parse(actual_end).date()

        try:
            project_id = json_data['projectId']
            related_project = models.Project.objects.get(pk=project_id)
            setattr(instance, 'project', related_project)
        except KeyError:
            raise InvalidObjectException(
                'Project phase {} has no projectId key to find its target'
                '- skipping.'.format(instance.id)
            )
        except ObjectDoesNotExist as e:
            raise InvalidObjectException(
                'Project phase {} has a projectId that does not exist.'
                ' ObjectDoesNotExist Exception: {}'.format(instance.id, e)
            )

        self.set_relations(instance, json_data)
        return instance

    def client_call(self, project_id, *args, **kwargs):
        return self.client.get_project_phases(project_id, *args, **kwargs)

    def get_page(self, *args, **kwargs):
        records = []
        project_qs = models.Project.objects.all().order_by(self.lookup_key)

        for project_id in project_qs.values_list('id', flat=True):
            records += self.client_call(project_id, *args, **kwargs)

        return records


class ProjectTypeSynchronizer(Synchronizer):
    client_class = api.ProjectAPIClient
    model_class = models.ProjectTypeTracker

    def _assign_field_data(self, instance, json_data):
        instance.id = json_data['id']
        instance.name = json_data['name']
        instance.default_flag = json_data.get('defaultFlag')
        instance.inactive_flag = json_data.get('inactiveFlag')
        return instance

    def get_page(self, *args, **kwargs):
        return self.client.get_project_types(*args, **kwargs)


class ProjectTeamMemberSynchronizer(Synchronizer):
    client_class = api.ProjectAPIClient
    model_class = models.ProjectTeamMemberTracker

    related_meta = {
        'member': (models.Member, 'member'),
        'workRole': (models.WorkRole, 'work_role')
    }

    def _assign_field_data(self, instance, json_data):
        instance.id = json_data['id']
        start_date = json_data.get('startDate')
        end_date = json_data.get('endDate')

        if start_date:
            instance.start_date = parse(start_date)
        if end_date:
            instance.end_date = parse(end_date)

        try:
            project_id = json_data['projectId']
            related_project = models.Project.objects.get(pk=project_id)
            setattr(instance, 'project', related_project)
        except KeyError:
            raise InvalidObjectException(
                'Project team member {} has no projectId key to find its '
                'target - skipping.'.format(instance.id)
            )
        except ObjectDoesNotExist as e:
            raise InvalidObjectException(
                'Project team member {} has a projectId that does not exist.'
                ' ObjectDoesNotExist Exception: {}'.format(instance.id, e)
            )

        self.set_relations(instance, json_data)
        return instance

    def client_call(self, project_id, *args, **kwargs):
        return self.client.get_project_team_members(
            project_id, *args, **kwargs)

    def get_page(self, *args, **kwargs):
        records = []
        project_qs = models.Project.objects.filter(
            status__closed_flag=False).order_by(self.lookup_key)

        for project_id in project_qs.values_list('id', flat=True):
            records += self.client_call(project_id, *args, **kwargs)

        return records


class ProjectSynchronizer(Synchronizer):
    client_class = api.ProjectAPIClient
    model_class = models.ProjectTracker
    related_meta = {
        'status': (models.ProjectStatus, 'status'),
        'manager': (models.Member, 'manager'),
        'company': (models.Company, 'company'),
        'type': (models.ProjectType, 'type'),
        'board': (models.ConnectWiseBoard, 'board'),
    }

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # Only sync projects in non-closed statuses. We could simply use
        # closedFlag=False but API versions before 2019.5 don't support the
        # closedFlag field and we need to support those versions for now.
        self.api_conditions = ['status/id in ({})'.format(
            ','.join(
                str(i.id) for
                i in models.ProjectStatus.objects.filter(closed_flag=False)
            )
        )]

    def _assign_field_data(self, instance, json_data):
        actual_start = json_data.get('actualStart')
        actual_end = json_data.get('actualEnd')
        estimated_start = json_data.get('estimatedStart')
        estimated_end = json_data.get('estimatedEnd')
        scheduled_start = json_data.get('scheduledStart')
        scheduled_end = json_data.get('scheduledEnd')

        instance.id = json_data['id']
        instance.name = json_data['name']
        instance.description = json_data.get('description')
        actual_hours = json_data.get('actualHours')
        budget_hours = json_data.get('budgetHours')
        scheduled_hours = json_data.get('scheduledHours')
        percent_complete = json_data.get('percentComplete')
        instance.udf = {str(item['id']): item
                        for item in json_data.get('customFields', list())}

        instance.actual_hours = Decimal(str(actual_hours)) \
            if actual_hours is not None else None
        instance.budget_hours = Decimal(str(budget_hours)) \
            if budget_hours is not None else None
        instance.scheduled_hours = Decimal(str(scheduled_hours)) \
            if scheduled_hours is not None else None
        instance.percent_complete = Decimal(str(percent_complete)) \
            if percent_complete is not None else None

        if actual_start:
            instance.actual_start = parse(actual_start).date()

        if actual_end:
            instance.actual_end = parse(actual_end).date()

        if estimated_start:
            instance.estimated_start = parse(estimated_start).date()

        if estimated_end:
            instance.estimated_end = parse(estimated_end).date()

        if scheduled_start:
            instance.scheduled_start = parse(scheduled_start).date()

        if scheduled_end:
            instance.scheduled_end = parse(scheduled_end).date()

        self.set_relations(instance, json_data)
        return instance

    def get_page(self, *args, **kwargs):
        return self.client.get_projects(*args, **kwargs)

    def get_single(self, project_id):
        return self.client.get_project(project_id)


class MemberSynchronizer(Synchronizer):
    client_class = api.SystemAPIClient
    model_class = models.MemberTracker

    related_meta = {
        'workRole': (models.WorkRole, 'work_role'),
        'workType': (models.WorkType, 'work_type')
    }

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.last_sync_job_time = None

    def _assign_field_data(self, instance, json_data):
        instance.id = json_data['id']
        instance.first_name = json_data.get('firstName')
        instance.last_name = json_data.get('lastName')
        instance.identifier = json_data.get('identifier')
        instance.office_email = json_data.get('officeEmail')
        instance.license_class = json_data.get('licenseClass')
        instance.inactive = json_data.get('inactiveFlag')

        self.set_relations(instance, json_data)

        return instance

    def _save_avatar(self, member, avatar, attachment_filename):
        """
        Thumbnails will only be generated if the filename hashed from the
        avatar is not the same as the currently saved avatar or the avatar
        does not exist in storage already.
        """
        thumbnail_size = {
            'avatar': (80, 80),
            'micro_avatar': (20, 20),
        }
        extension = get_filename_extension(attachment_filename)
        filename = '{}.{}'.format(get_hash(avatar), extension)

        current_avatar = member.avatar
        current_avatar_filename = '{}20x20.{}'.format(member.avatar, extension)
        process_thumbnails = filename != current_avatar

        if process_thumbnails or \
                not default_storage.exists(current_avatar_filename):
            member.avatar = filename
            avatar_file = None
            for size in thumbnail_size:
                current_filename = generate_filename(thumbnail_size[size],
                                                     current_avatar, extension)
                filename = '{}.{}'.format(get_hash(avatar), extension)
                try:
                    avatar_file, filename = generate_thumbnail(
                        avatar, thumbnail_size[size],
                        extension, filename)
                except Exception as e:
                    logger.warning("Error saving member avatar. {}".format(e))

                # Just delete the currently saved avatar image
                # so we don't need to worry about cleaning up
                # duplicates or old thumbnails
                if avatar_file and filename:
                    default_storage.delete(current_filename)

                    logger.info("Saving member '{}' avatar to {}.".format(
                        member.identifier, filename))
                    default_storage.save(filename, avatar_file)

    def get_page(self, *args, **kwargs):
        return self.client.get_members(*args, **kwargs)

    def update_or_create_instance(self, api_instance):
        """
        In addition to what the parent does, also update avatar if necessary.
        """
        try:
            instance, result = super().update_or_create_instance(api_instance)
        except IntegrityError as e:
            raise InvalidObjectException(
                'Failed to update member: {}'.format(e)
            )
        username = instance.identifier
        photo_id = None

        # Only update the avatar if the member profile
        # was updated since last sync.
        member_last_updated = parse(api_instance['_info']['lastUpdated'])
        member_stale = False
        if self.last_sync_job_time:
            member_stale = member_last_updated > self.last_sync_job_time

        if api_instance.get('photo'):
            photo_id = api_instance['photo']['id']

        # For when a CW user removes their photo without setting
        # a new one. Remove the image from storage.
        if instance.avatar and not photo_id:
            try:
                remove_thumbnail(instance.avatar)
            except NoCredentialsError as e:
                msg = 'Error when removing thumbnail for {}.' \
                      ' NoCredentialsError Exception: {}.' \
                    .format(instance.username, e)
                logger.warning(msg)
            instance.avatar = None
            instance.save()
        # Fetch the image when:
        # CW tells us where the image is AND one of the following is true:
        # * this is a full sync
        # * there's no previous member sync job record
        # * the member's avatar doesn't already exist
        # * our member record is stale
        # * the member was created
        if photo_id and (self.full or
                         not self.last_sync_job_time or
                         not bool(instance.avatar) or
                         member_stale or
                         result == CREATED):
            logger.info(
                'Fetching avatar for member {}.'.format(username)
            )
            (attachment_filename, avatar) = self.client \
                .get_member_image_by_photo_id(photo_id, username)
            if attachment_filename and avatar:
                try:
                    self._save_avatar(instance, avatar, attachment_filename)
                except NoCredentialsError as e:
                    msg = 'Error when saving avatar for {}.' \
                          ' NoCredentialsError Exception: {}.' \
                          .format(username, e)
                    logger.warning(msg)
            instance.save()

        return instance, result


class TicketSynchronizerMixin:
    model_class = models.TicketTracker
    batch_condition_list = []

    related_meta = {
        'team': (models.Team, 'team'),
        'board': (models.ConnectWiseBoard, 'board'),
        'company': (models.Company, 'company'),
        'priority': (models.TicketPriority, 'priority'),
        'project': (models.Project, 'project'),
        'phase': (models.ProjectPhase, 'phase'),
        'serviceLocation': (models.Location, 'location'),
        'status': (models.BoardStatus, 'status'),
        'owner': (models.Member, 'owner'),
        'sla': (models.Sla, 'sla'),
        'type': (models.Type, 'type'),
        'subType': (models.SubType, 'sub_type'),
        'item': (models.Item, 'sub_type_item'),
        'agreement': (models.Agreement, 'agreement'),
    }

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.api_conditions = ['closedFlag=False']
        # To get all open tickets, we can simply supply a `closedFlag=False`
        # condition for on-premise ConnectWise. But for hosted ConnectWise,
        # this results in timeouts for requests, so we also need to add a
        # condition for all the open statuses. This doesn't impact on-premise
        # ConnectWise, so we just do it for all cases.
        request_settings = DjconnectwiseSettings().get_settings()
        board_names = request_settings.get('board_status_filter')
        filtered_statuses = models.BoardStatus.available_objects.filter(
            closed_status=False).order_by(self.lookup_key)

        if board_names:
            boards = [board.strip() for board in board_names.split(',')]

            boards_exist = models.ConnectWiseBoard.available_objects.filter(
                name__in=boards).exists()

            if boards_exist:
                filtered_statuses = filtered_statuses.filter(
                    board__name__in=boards
                )

        if filtered_statuses:
            # Only do this if we know of at least one open status.
            self.batch_condition_list = \
                list(filtered_statuses.values_list('id', flat=True))

    def get_batch_condition(self, conditions):
        batch_condition = 'status/id in ({})'.format(
            ','.join([str(i) for i in conditions])
        )

        request_settings = DjconnectwiseSettings().get_settings()
        keep_closed = request_settings.get('keep_closed_ticket_days')
        if keep_closed:
            batch_condition = self.format_conditions(
                keep_closed, batch_condition, request_settings)

        return batch_condition

    def format_conditions(self, keep_closed,
                          batch_condition, request_settings):
        closed_date = timezone.now() - timezone.timedelta(days=keep_closed)
        condition = 'closedDate>[{}]'.format(closed_date)

        keep_closed_board_ids = \
            request_settings.get('keep_closed_status_board_ids')
        if keep_closed_board_ids:
            condition = '{} and board/id in ({})'.format(
                condition, keep_closed_board_ids)

        # lastUpdated is only present when running full syncs.
        if not self.full:
            last_updated = self.api_conditions[-1]
            condition = '{} and {}'.format(condition, last_updated)

        batch_condition = '{} or {}'.format(batch_condition, condition)
        return batch_condition

    def get_page(self, *args, **kwargs):
        return self.client.get_tickets(*args, **kwargs)

    def get_single(self, ticket_id):
        return self.client.get_ticket(ticket_id)

    # sync_config is read-only, so mutable argument is used for simplicity.
    def sync_related(self, instance, sync_config={}):
        instance_id = instance.id
        sync_classes = []

        if sync_config.get('sync_service_note', True):
            note_sync = ServiceNoteSynchronizer()
            note_sync.api_conditions = [instance_id]
            sync_classes.append((note_sync, Q(ticket=instance)))

        if sync_config.get('sync_time_entry', True):
            time_sync = TimeEntrySynchronizer()
            time_sync.batch_condition_list = [instance_id]
            sync_classes.append((time_sync, Q(charge_to_id=instance)))

        if sync_config.get('sync_activity', True):
            activity_sync = ActivitySynchronizer()
            activity_sync.api_conditions = ['ticket/id={}'.format(instance_id)]
            sync_classes.append((activity_sync, Q(ticket=instance)))

        self.sync_children(*sync_classes)

    def fetch_sync_by_id(self, instance_id, sync_config={}):
        instance = super().fetch_sync_by_id(instance_id)
        if not instance.closed_flag:
            self.sync_related(instance, sync_config)
        return instance

    def _instance_ids(self, filter_params=None):
        tickets_qset = self.filter_by_record_type().order_by(self.lookup_key)

        if not filter_params:
            ids = tickets_qset.values_list(self.lookup_key, flat=True)
        else:
            ids = tickets_qset.filter(filter_params).values_list(
                self.lookup_key, flat=True
            )
        return set(ids)

    def get_delete_qset(self, stale_ids):
        tickets = self.filter_by_record_type()
        return tickets.filter(pk__in=stale_ids)

    def get_sync_job_qset(self):
        return models.SyncJob.objects.filter(
            entity_name=self.model_class.__bases__[0].__name__,
            synchronizer_class=self.__class__.__name__
        )

    def _assign_field_data(self, instance, json_data):

        instance.id = json_data['id']
        instance.summary = json_data['summary']
        instance.closed_flag = json_data.get('closedFlag')
        instance.last_updated_utc = json_data.get('_info').get('lastUpdated')
        instance.required_date_utc = json_data.get('requiredDate')
        instance.resources = json_data.get('resources')
        instance.bill_time = json_data.get('billTime')
        instance.customer_updated = json_data.get('customerUpdatedFlag')

        # Key is comes out of db as string, so we add it as a string here
        # so the tracker can compare it properly.
        # TODO add udf to other models
        instance.udf = {str(item['id']): item
                        for item in json_data.get('customFields', list())}

        instance.automatic_email_cc_flag = \
            json_data.get('automaticEmailCcFlag', False)
        instance.automatic_email_contact_flag = \
            json_data.get('automaticEmailContactFlag', False)
        instance.automatic_email_resource_flag = \
            json_data.get('automaticEmailResourceFlag', False)
        instance.automatic_email_cc = \
            json_data.get('automaticEmailCc')
        if instance.automatic_email_cc:
            # Truncate the field to 1000 characters as per CW docs for the
            # automatic_email_cc field, because in some cases more can be
            # received which causes a DataError. It is preferred to keep the
            # DB schema in-line with the CW specifications, even if the
            # specifications are wrong.
            instance.automatic_email_cc = instance.automatic_email_cc[:1000]

        budget_hours = json_data.get('budgetHours')
        actual_hours = json_data.get('actualHours')
        instance.budget_hours = Decimal(str(budget_hours)) \
            if budget_hours is not None else None
        instance.actual_hours = Decimal(str(actual_hours)) \
            if actual_hours is not None else None

        instance.predecessor_type = json_data.get('predecessorType')

        instance.lag_days = json_data.get('lagDays')
        instance.lag_non_working_days_flag = \
            json_data.get('lagNonworkingDaysFlag', False)

        try:
            predecessor_id = json_data.get('predecessorId')

            if predecessor_id:
                if instance.predecessor_type == self.model_class.TICKET:
                    instance.ticket_predecessor = \
                        models.Ticket.objects.get(pk=predecessor_id)
                else:
                    instance.phase_predecessor = \
                        models.ProjectPhase.objects.get(pk=predecessor_id)

        except ObjectDoesNotExist as e:
            logger.warning(
                'Ticket {} has a predecessorId that does not exist. '
                'ObjectDoesNotExist Exception: {}'.format(instance.id, e)
            )

        return instance


class ServiceTicketSynchronizer(TicketSynchronizerMixin,
                                BatchConditionMixin, Synchronizer):
    client_class = api.ServiceAPIClient

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        request_settings = DjconnectwiseSettings().get_settings()

        sync_child_tickets = request_settings.get('sync_child_tickets')
        if not sync_child_tickets and sync_child_tickets is not None:
            self.api_conditions.append('parentTicketId=null')

    def _assign_field_data(self, instance, json_data):
        instance = super()._assign_field_data(instance, json_data)

        entered_date_utc = json_data.get('_info').get('dateEntered')

        instance.record_type = json_data.get('recordType')
        instance.parent_ticket_id = json_data.get('parentTicketId')
        instance.has_child_ticket = json_data.get('hasChildTicket')
        instance.respond_mins = json_data.get('respondMinutes')
        instance.res_plan_mins = json_data.get('resPlanMinutes')
        instance.resolve_mins = json_data.get('resolveMinutes')
        instance.date_resolved_utc = json_data.get('dateResolved')
        instance.date_resplan_utc = json_data.get('dateResplan')
        instance.date_responded_utc = json_data.get('dateResponded')

        if entered_date_utc:
            # Parse the date here so that a datetime object is
            # available for SLA calculation.
            instance.entered_date_utc = parse(entered_date_utc)
        if instance.last_updated_utc:
            instance.last_updated_utc = parse(instance.last_updated_utc)
        if instance.required_date_utc:
            instance.required_date_utc = parse(instance.required_date_utc)
        if instance.date_resolved_utc:
            instance.date_resolved_utc = parse(instance.date_resolved_utc)
        if instance.date_resplan_utc:
            instance.date_resplan_utc = parse(instance.date_resplan_utc)
        if instance.date_responded_utc:
            instance.date_responded_utc = parse(instance.date_responded_utc)

        self.set_relations(instance, json_data)
        return instance

    def filter_by_record_type(self):
        return self.model_class.objects.filter(
            record_type=models.Ticket.SERVICE_TICKET)


class ProjectTicketSynchronizer(TicketSynchronizerMixin,
                                BatchConditionMixin, Synchronizer):
    client_class = api.ProjectAPIClient

    def _assign_field_data(self, instance, json_data):
        instance = super()._assign_field_data(instance, json_data)

        if instance.last_updated_utc:
            instance.last_updated_utc = parse(instance.last_updated_utc)
        if instance.required_date_utc:
            instance.required_date_utc = parse(instance.required_date_utc)

        # Tickets from the project/tickets API endpoint do not include the
        # record type field but we use the same Ticket model for project and
        # service tickets so we need to set the record type here.
        if json_data.get('isIssueFlag'):
            instance.record_type = models.Ticket.PROJECT_ISSUE
        else:
            instance.record_type = models.Ticket.PROJECT_TICKET

        self.set_relations(instance, json_data)
        return instance

    def filter_by_record_type(self):
        project_record_types = [
            models.Ticket.PROJECT_ISSUE, models.Ticket.PROJECT_TICKET
        ]
        return self.model_class.objects.filter(
            record_type__in=project_record_types)


class CalendarSynchronizer(Synchronizer):
    client_class = api.ScheduleAPIClient
    model_class = models.CalendarTracker

    def _assign_field_data(self, instance, json_data):
        instance.id = json_data['id']
        instance.name = json_data['name']
        instance.monday_start_time = \
            parse(json_data.get('mondayStartTime')).time() \
            if json_data.get('mondayStartTime') else None
        instance.monday_end_time = \
            parse(json_data.get('mondayEndTime')).time() \
            if json_data.get('mondayEndTime') else None
        instance.tuesday_start_time = \
            parse(json_data.get('tuesdayStartTime')).time() \
            if json_data.get('tuesdayStartTime') else None
        instance.tuesday_end_time = \
            parse(json_data.get('tuesdayEndTime')).time() \
            if json_data.get('tuesdayEndTime') else None
        instance.wednesday_start_time = \
            parse(json_data.get('wednesdayStartTime')).time() \
            if json_data.get('wednesdayStartTime') else None
        instance.wednesday_end_time = \
            parse(json_data.get('wednesdayEndTime')).time() \
            if json_data.get('wednesdayEndTime') else None
        instance.thursday_start_time = \
            parse(json_data.get('thursdayStartTime')).time() \
            if json_data.get('thursdayStartTime') else None
        instance.thursday_end_time = \
            parse(json_data.get('thursdayEndTime')).time() \
            if json_data.get('thursdayEndTime') else None
        instance.friday_start_time = \
            parse(json_data.get('fridayStartTime')).time() \
            if json_data.get('fridayStartTime') else None
        instance.friday_end_time = \
            parse(json_data.get('fridayEndTime')).time() \
            if json_data.get('fridayEndTime') else None
        instance.saturday_start_time = \
            parse(json_data.get('saturdayStartTime')).time() \
            if json_data.get('saturdayStartTime') else None
        instance.saturday_end_time = \
            parse(json_data.get('saturdayEndTime')).time() \
            if json_data.get('saturdayEndTime') else None
        instance.sunday_start_time = \
            parse(json_data.get('sundayStartTime')).time() \
            if json_data.get('sundayStartTime') else None
        instance.sunday_end_time = \
            parse(json_data.get('sundayEndTime')).time() \
            if json_data.get('sundayEndTime') else None

        self._assign_relation(
            instance,
            json_data,
            'holidayList',
            models.HolidayList,
            'holiday_list'
            )

        return instance

    def get_page(self, *args, **kwargs):
        return self.client.get_calendars(*args, **kwargs)


class HolidaySynchronizer(Synchronizer):
    client_class = api.ScheduleAPIClient
    model_class = models.HolidayTracker

    def _assign_field_data(self, instance, json_data):

        instance.id = json_data['id']
        instance.name = json_data['name']
        instance.all_day_flag = json_data.get('allDayFlag')
        instance.date = parse(json_data.get('date')).date()
        instance.start_time = \
            parse(json_data.get('timeStart')).time() \
            if json_data.get('timeStart') else None
        instance.end_time = \
            parse(json_data.get('timeEnd')).time() \
            if json_data.get('timeEnd') else None

        self._assign_relation(
            instance,
            json_data,
            'holidayList',
            models.HolidayList,
            'holiday_list'
            )

        return instance

    def client_call(self, list_id, *args, **kwargs):
        return self.client.get_holidays(list_id, *args, **kwargs)

    def get_page(self, *args, **kwargs):
        records = []
        list_qs = models.HolidayList.objects.all()

        for list_id in list_qs.values_list('id', flat=True):
            records += self.client_call(list_id, *args, **kwargs)
        return records


class HolidayListSynchronizer(Synchronizer):
    client_class = api.ScheduleAPIClient
    model_class = models.HolidayListTracker

    def _assign_field_data(self, instance, json_data):
        instance.id = json_data['id']
        instance.name = json_data['name']

        return instance

    def get_page(self, *args, **kwargs):
        return self.client.get_holiday_lists(*args, **kwargs)


class SLAPrioritySynchronizer(Synchronizer):
    client_class = api.ServiceAPIClient
    model_class = models.SlaPriorityTracker

    related_meta = {
        'priority': (models.TicketPriority, 'priority'),
        'sla': (models.Sla, 'sla')
    }

    def _assign_field_data(self, instance, json_data):
        instance.id = json_data['id']
        instance.respond_hours = json_data['respondHours']
        instance.plan_within = json_data['planWithin']
        instance.resolution_hours = json_data['resolutionHours']

        self.set_relations(instance, json_data)
        return instance

    def client_call(self, sla_id, *args, **kwargs):
        return self.client.get_slapriorities(sla_id, *args, **kwargs)

    def get_page(self, *args, **kwargs):
        records = []
        sla_qs = models.Sla.objects.all()

        for sla_id in sla_qs.values_list('id', flat=True):
            records += self.client_call(sla_id, *args, **kwargs)

        return records


class SLASynchronizer(Synchronizer):
    client_class = api.ServiceAPIClient
    model_class = models.SlaTracker

    related_meta = {
        'customCalendar': (models.Calendar, 'calendar'),
    }

    def _assign_field_data(self, instance, json_data):
        instance.id = json_data['id']
        instance.name = json_data['name']
        instance.default_flag = json_data['defaultFlag']
        instance.respond_hours = json_data['respondHours']
        instance.plan_within = json_data['planWithin']
        instance.resolution_hours = json_data['resolutionHours']
        instance.based_on = json_data.get('basedOn')

        self.set_relations(instance, json_data)
        return instance

    def get_page(self, *args, **kwargs):
        return self.client.get_slas(*args, **kwargs)


class OpportunitySynchronizer(Synchronizer):
    client_class = api.SalesAPIClient
    model_class = models.OpportunityTracker
    related_meta = {
        'type': (models.OpportunityType, 'opportunity_type'),
        'stage': (models.OpportunityStage, 'stage'),
        'status': (models.OpportunityStatus, 'status'),
        'probability': (models.SalesProbability, 'probability'),
        'primarySalesRep': (models.Member, 'primary_sales_rep'),
        'secondarySalesRep': (models.Member, 'secondary_sales_rep'),
        'company': (models.Company, 'company'),
        'closedBy': (models.Member, 'closed_by')
    }

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # only synchronize Opportunities that do not have an OpportunityStatus
        # with the closedFlag=True
        open_statuses = list(
            models.OpportunityStatus.objects.
            filter(closed_flag=False).
            order_by(self.lookup_key).
            values_list('id', flat=True)
        )
        if open_statuses:
            # Only do this if we know of at least one open status.
            self.api_conditions.append(
                'status/id in ({})'.format(
                    ','.join([str(i) for i in open_statuses])
                )
            )

    def _update_or_create_child(self, model_class, json_data):
        child_name = json_data['name']
        # Setting the name default ensures that if Django has to create the
        # object, then the name has already been set and we don't have to save
        # again.
        child, created = model_class.objects.get_or_create(
            id=json_data['id'],
            defaults={
                'name': child_name,
            }
        )
        if not created:
            # Ensure the name is up to date.
            child.name = child_name

            # Don't save it if there was no change
            if child.tracker.changed():
                child.save()
        return child

    def _assign_field_data(self, instance, json_data):
        instance.id = json_data['id']
        instance.name = json_data['name']
        instance.notes = json_data.get('notes')
        instance.source = json_data.get('source')
        instance.location_id = json_data.get('locationId')
        instance.business_unit_id = json_data.get('businessUnitId')
        instance.customer_po = json_data.get('customerPO')
        instance.udf = {str(item['id']): item
                        for item in json_data.get('customFields', list())}

        # handle dates
        expected_close_date = json_data.get('expectedCloseDate')
        if expected_close_date:
            instance.expected_close_date = parse(expected_close_date).date()

        pipeline_change_date = json_data.get('pipelineChangeDate')
        if pipeline_change_date:
            instance.pipeline_change_date = parse(pipeline_change_date)

        date_became_lead = json_data.get('dateBecameLead')
        if date_became_lead:
            instance.date_became_lead = parse(date_became_lead)

        closed_date = json_data.get('closedDate')
        if closed_date:
            instance.closed_date = parse(closed_date)

        priority = json_data.get('priority')
        if priority:
            instance.priority = self._update_or_create_child(
                models.OpportunityPriorityTracker, priority
            )

        self.set_relations(instance, json_data)
        return instance

    def get_page(self, *args, **kwargs):
        return self.client.get_opportunities(*args, **kwargs)

    def get_single(self, opportunity_id):
        return self.client.by_id(opportunity_id)


class OpportunityStageSynchronizer(Synchronizer):
    client_class = api.SalesAPIClient
    model_class = models.OpportunityStageTracker

    def _assign_field_data(self, instance, json_data):
        instance.id = json_data['id']
        instance.name = json_data['name']

        return instance

    def get_page(self, *args, **kwargs):
        return self.client.get_opportunity_stages(*args, **kwargs)


class OpportunityStatusSynchronizer(Synchronizer):
    client_class = api.SalesAPIClient
    model_class = models.OpportunityStatusTracker

    def _assign_field_data(self, instance, json_data):
        instance.id = json_data['id']
        instance.name = json_data['name']
        instance.won_flag = json_data['wonFlag']
        instance.lost_flag = json_data['lostFlag']
        instance.closed_flag = json_data['closedFlag']
        instance.inactive_flag = json_data['inactiveFlag']
        return instance

    def get_page(self, *args, **kwargs):
        return self.client.get_opportunity_statuses(*args, **kwargs)


class OpportunityTypeSynchronizer(Synchronizer):
    client_class = api.SalesAPIClient
    model_class = models.OpportunityTypeTracker

    def _assign_field_data(self, instance, json_data):
        instance.id = json_data['id']
        instance.description = json_data['description']
        instance.inactive_flag = json_data['inactiveFlag']
        return instance

    def get_page(self, *args, **kwargs):
        return self.client.get_opportunity_types(*args, **kwargs)


class TypeSynchronizer(BoardFilterMixin, Synchronizer):
    client_class = api.ServiceAPIClient
    model_class = models.TypeTracker

    related_meta = {
        'board': (models.ConnectWiseBoard, 'board')
    }

    def _assign_field_data(self, instance, json_data):
        instance.id = json_data['id']
        instance.name = json_data['name']

        self.set_relations(instance, json_data)
        return instance

    def client_call(self, board_id, *args, **kwargs):
        return self.client.get_types(board_id, *args, **kwargs)


class SubTypeSynchronizer(BoardFilterMixin, Synchronizer):
    client_class = api.ServiceAPIClient
    model_class = models.SubTypeTracker

    related_meta = {
        'board': (models.ConnectWiseBoard, 'board')
    }

    def _assign_field_data(self, instance, json_data):
        instance.id = json_data['id']
        instance.name = json_data['name']

        self.set_relations(instance, json_data)
        return instance

    def client_call(self, board_id, *args, **kwargs):
        return self.client.get_subtypes(board_id, *args, **kwargs)


class ItemSynchronizer(BoardFilterMixin, Synchronizer):
    client_class = api.ServiceAPIClient
    model_class = models.ItemTracker

    related_meta = {
        'board': (models.ConnectWiseBoard, 'board')
    }

    def _assign_field_data(self, instance, json_data):
        instance.id = json_data['id']
        instance.name = json_data['name']

        self.set_relations(instance, json_data)
        return instance

    def client_call(self, board_id, *args, **kwargs):
        return self.client.get_items(board_id, *args, **kwargs)


class WorkTypeSynchronizer(Synchronizer):
    client_class = api.TimeAPIClient
    model_class = models.WorkTypeTracker

    def _assign_field_data(self, instance, json_data):
        instance.id = json_data['id']
        instance.name = json_data['name']
        instance.inactive_flag = json_data['inactiveFlag']
        instance.overall_default_flag = json_data['overallDefaultFlag']
        if json_data['billTime'] == 'NoDefault':
            instance.bill_time = None
        else:
            instance.bill_time = json_data['billTime']

        return instance

    def get_page(self, *args, **kwargs):
        return self.client.get_work_types(*args, **kwargs)


class WorkRoleSynchronizer(Synchronizer):
    client_class = api.TimeAPIClient
    model_class = models.WorkRoleTracker

    def _assign_field_data(self, instance, json_data):
        instance.id = json_data['id']
        instance.name = json_data['name']
        instance.inactive_flag = json_data['inactiveFlag']

        return instance

    def get_page(self, *args, **kwargs):
        return self.client.get_work_roles(*args, **kwargs)


class AgreementSynchronizer(Synchronizer):
    client_class = api.FinanceAPIClient
    model_class = models.AgreementTracker

    related_meta = {
        'workRole': (models.WorkRole, 'work_role'),
        'workType': (models.WorkType, 'work_type'),
        'company': (models.Company, 'company')
    }

    def _assign_field_data(self, instance, json_data):
        instance.id = json_data['id']
        instance.name = json_data['name']
        instance.agreement_type = json_data['type']['name']
        instance.agreement_status = json_data.get('agreementStatus')
        instance.cancelled_flag = json_data['cancelledFlag']
        if json_data['billTime'] == 'NoDefault':
            instance.bill_time = None
        else:
            instance.bill_time = json_data['billTime']

        self.set_relations(instance, json_data)
        return instance

    def get_page(self, *args, **kwargs):
        return self.client.get_agreements(*args, **kwargs)


class UDFSynchronizer(Synchronizer):

    def fetch_records(self, results, conditions=None):
        """
        Only fetch one record. To ensure we get a record if one exists, just
        make a call for tickets but only ask for 1 pages, with a page size of 1
        """
        page = 1
        logger.info('Fetching {} records'.format(
            self.model_class.__bases__[0].__name__))
        page_records = self.get_page(
            page=page,
            page_size=page,
        )
        for record in page_records:
            # Should only run once, or not at all if there are 0 records of
            # requested type.
            self.persist_page(record.get('customFields', list()), results)

        return results

    def _assign_field_data(self, instance, json_data):
        instance.id = json_data['id']
        instance.caption = json_data.get('caption')
        instance.type = json_data.get('type')
        instance.entry_method = json_data.get('entryMethod')
        instance.number_of_decimals = json_data.get('numberOfDecimals')

        return instance


class TicketUDFSynchronizer(UDFSynchronizer):
    client_class = api.ServiceAPIClient
    model_class = models.TicketUDFTracker

    def get_page(self, *args, **kwargs):
        # Using the Service API Client fine for both service and project
        #   tickets as both ticket types have all UDFs returned in a request
        return self.client.get_tickets(*args, **kwargs)


class ProjectUDFSynchronizer(UDFSynchronizer):
    client_class = api.ProjectAPIClient
    model_class = models.ProjectUDFTracker

    def get_page(self, *args, **kwargs):
        return self.client.get_projects(*args, **kwargs)


class ActivityUDFSynchronizer(UDFSynchronizer):
    client_class = api.SalesAPIClient
    model_class = models.ActivityUDFTracker

    def get_page(self, *args, **kwargs):
        return self.client.get_activities(*args, **kwargs)


class OpportunityUDFSynchronizer(UDFSynchronizer):
    client_class = api.SalesAPIClient
    model_class = models.OpportunityUDFTracker

    def get_page(self, *args, **kwargs):
        return self.client.get_opportunities(*args, **kwargs)



from django.core.management.base import BaseCommand, CommandError
from djconnectwise.callbacks import CallbacksHandler
from djconnectwise.api import ConnectWiseAPIError


class Command(BaseCommand):
    help = 'Ensure callbacks are registered on ConnectWise.'

    def handle(self, *args, **options):
        handler = CallbacksHandler()
        try:
            handler.ensure_registered()
        except ConnectWiseAPIError as e:
            raise CommandError(e)



# Generated by Django 2.1 on 2019-07-16 16:30

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('djconnectwise', '0092_remove_ticket_work_type'),
    ]

    operations = [
        migrations.AddField(
            model_name='ticket',
            name='automatic_email_cc',
            field=models.CharField(blank=True, max_length=1000, null=True),
        ),
        migrations.AddField(
            model_name='timeentry',
            name='email_cc',
            field=models.CharField(blank=True, max_length=1000, null=True),
        ),
    ]



from django.urls import re_path
from . import views

app_name = 'djconnectwise'

urlpatterns = [
    re_path(
        r'^callback/$',
        view=views.CallBackView.as_view(),
        name='callback'
    ),
]



import datetime
from unittest.mock import patch

from django.utils import timezone
from freezegun import freeze_time
from model_mommy import mommy
from test_plus.test import TestCase

from djconnectwise.models import MyCompanyOther, Holiday, HolidayList, \
    Ticket, InvalidStatusError, Sla, Calendar, SlaPriority, TicketPriority, \
    BoardStatus

ticket_statuses_names = [
    'New',
    'In Progress',
    'Scheduled',
    'Blocked',
    'Completed',
    'Waiting For Client',
    'Closed',
]
escalation_stages = [
    'NotResponded',
    'ResolutionPlan',
    'Responded',
    'NoEscalation',
    'Resolved',
    'NoEscalation',
    'Resolved'
]


class ModelTestCase(TestCase):

    def setUp(self):
        self.members = mommy.make_recipe(
            'djconnectwise.tests.member',
            _quantity=3
        )
        self.projects = mommy.make_recipe(
            'djconnectwise.tests.project',
            _quantity=4,
        )
        self.companies = mommy.make_recipe(
            'djconnectwise.tests.company',
            _quantity=2
        )
        self.ticket_priorities = mommy.make_recipe(
            'djconnectwise.tests.ticket_priority',
            _quantity=3
        )
        self.connectwise_boards = mommy.make_recipe(
            'djconnectwise.tests.connectwise_board',
            _quantity=3,
        )
        sla = Sla.objects.create(
            name="Standard SLA",
            default_flag=True,
            respond_hours=2,
            plan_within=8,
            resolution_hours=96,
            based_on='MyCalendar'
        )

        goal_times = [
            (2.0, 4.0, 8.0),
            (6.0, 12.0, 72.0),
        ]
        sla_priorities = []
        # Don't set an SLA priority for the first ticket priority to ensure
        # it uses the default SLA response goal hours.
        for index, ticket_priority in enumerate(self.ticket_priorities[-2:]):
            sla_priority = SlaPriority.objects.create(
                sla=sla,
                priority=ticket_priority,
                respond_hours=goal_times[index][0],
                plan_within=goal_times[index][1],
                resolution_hours=goal_times[index][2]
            )
            sla_priorities.append(sla_priority)

        holiday_list = HolidayList.objects.create(
            name="Test List"
        )
        Holiday.objects.create(
            name="Test Holiday",
            all_day_flag=True,
            date=datetime.datetime(year=2017, day=12, month=12),
            holiday_list=holiday_list
        )
        calendar = Calendar.objects.create(
            name="Standard Office Hours",
            monday_start_time='08:00:00',
            monday_end_time='17:00:00',
            tuesday_start_time='08:00:00',
            tuesday_end_time='17:00:00',
            wednesday_start_time='08:00:00',
            wednesday_end_time='17:00:00',
            thursday_start_time='08:00:00',
            thursday_end_time='17:00:00',
            friday_start_time='08:00:00',
            friday_end_time='17:00:00',
            saturday_start_time=None,
            saturday_end_time=None,
            sunday_start_time=None,
            sunday_end_time=None,
            holiday_list=holiday_list
        )
        MyCompanyOther.objects.create(
            default_calendar=calendar
        )
        Calendar.objects.create()
        for i, s in enumerate(ticket_statuses_names):
            # All the statuses belong to the first board.
            BoardStatus.objects.create(
                name=s,
                sort_order=i,
                board=self.connectwise_boards[0],
                closed_status=True if s in ['Completed', 'Closed'] else False,
                escalation_status=escalation_stages[i],
                display_on_board=True,
                inactive=False,
            )


class TestTicketPriority(TestCase):

    def test_priority_color(self):
        # If a priority has a color, then the property returns it.
        p = TicketPriority(name='Priority 1 - Emergency Response')
        # This also tests the setter.
        p.color = 'PapayaWhip'  # Yeah, PapayaWhip is a CSS color
        self.assertEqual(p.color, 'PapayaWhip')

    def test_priority_color_property(self):
        # If a priority doesn't have a color, then the property returns
        # a sensible default.
        p = TicketPriority(name='Priority 1 - Emergency Response')
        self.assertEqual(p.color, 'red')
        p = TicketPriority(name='Priority 2 - Quick Response')
        self.assertEqual(p.color, 'orange')
        p = TicketPriority(name='Priority 3 - Normal Response')
        self.assertEqual(p.color, 'yellow')
        p = TicketPriority(name='Priority 4 - Scheduled Maintenance')
        self.assertEqual(p.color, 'white')
        p = TicketPriority(name='Priority 5 - Next Time')
        self.assertEqual(p.color, 'darkmagenta')
        p = TicketPriority(name='Totally unknown priority')
        self.assertEqual(p.color, 'darkgray')


class TestBoard(ModelTestCase):
    def test_get_closed_status_prefers_closed(self):
        board = self.connectwise_boards[0]
        self.assertEqual(
            board.get_closed_status().name,
            'Closed'
        )

    def test_get_closed_status_any_closed(self):
        # Return some closed status if there's not a closed status called
        # 'Closed'
        board = self.connectwise_boards[0]
        board.board_statuses.filter(name='Closed').delete()
        closed_status = board.get_closed_status()
        self.assertTrue(closed_status.closed_status)
        self.assertNotEqual(
            closed_status.name,
            'Closed'
        )

    def test_get_closed_status_no_closed(self):
        # Return None if there's no closed statuses
        board = self.connectwise_boards[0]
        board.board_statuses.filter(closed_status=True).delete()
        closed_status = board.get_closed_status()
        self.assertEqual(closed_status, None)


class TestBoardStatus(ModelTestCase):
    def test_compare_lt(self):
        board = self.connectwise_boards[0]
        status_a = board.board_statuses[0]
        status_b = board.board_statuses[1]
        self.assertTrue(status_a < status_b)
        self.assertFalse(status_b < status_a)

    def test_compare_gt(self):
        board = self.connectwise_boards[0]
        status_a = board.board_statuses[0]
        status_b = board.board_statuses[1]
        self.assertTrue(status_b > status_a)
        self.assertFalse(status_a > status_b)

    def test_compare_none(self):
        board = self.connectwise_boards[0]
        status = board.board_statuses[0]
        self.assertFalse(None > status)
        self.assertFalse(status > None)
        self.assertFalse(None < status)
        self.assertFalse(status < None)


class TestCalendar(ModelTestCase):

    def test_get_first_day(self):
        calendar = Calendar.objects.first()

        day, days = calendar.get_first_day(
            datetime.datetime(year=2018, day=29, month=9))
        self.assertEqual(day, 0)
        self.assertEqual(days, 2)

        day, days = calendar.get_first_day(
            datetime.datetime(year=2018, day=19, month=9))
        self.assertEqual(day, 2)
        self.assertEqual(days, 0)

        day, days = calendar.get_first_day(
            datetime.datetime(year=2018, day=23, month=9))
        self.assertEqual(day, 0)
        self.assertEqual(days, 1)

    @freeze_time("2018-08-23 12:12:12", tz_offset=-7)
    def test_get_sla_time(self):
        # Ticket._get_sla_time(self, start, end, calendar)
        start = timezone.now()
        end = start + datetime.timedelta(days=1)
        calendar = Calendar.objects.first()

        sla_time = calendar.get_sla_time(start, end)
        self.assertEqual(sla_time, 540)

    @freeze_time("2018-08-23 12:12:12", tz_offset=-7)
    def test_get_sla_time_same_day(self):
        start = timezone.now()
        end = start + datetime.timedelta(hours=1)
        calendar = Calendar.objects.first()

        sla_time = calendar.get_sla_time(start, end)
        self.assertEqual(sla_time, 60)

    @freeze_time("2018-08-25 12:00:00", tz_offset=-7)
    def test_get_sla_time_over_weeked(self):
        start = timezone.now()
        end = start + datetime.timedelta(days=2)
        calendar = Calendar.objects.first()

        sla_time = calendar.get_sla_time(start, end)
        self.assertEqual(sla_time, 240)

    @freeze_time("2018-08-25 12:00:00", tz_offset=-7)
    def test_sla_get_time_same_day_weekend(self):
        start = timezone.now()
        end = start + datetime.timedelta(hours=2)
        calendar = Calendar.objects.first()

        sla_time = calendar.get_sla_time(start, end)
        self.assertEqual(sla_time, 0)


class TestTicket(ModelTestCase):

    def test_str(self):
        t = Ticket(id=1, summary='Únicôde wôrks!')
        self.assertEqual(
            '{}'.format(t),
            '1-Únicôde wôrks!'
        )

    def test_save_checks_status(self):
        # Raises an exception if the ticket status isn't valid for the
        # ticket's board.
        ticket = Ticket.objects.create(
            summary='test',
            status=self.connectwise_boards[0].board_statuses.first(),
            board=self.connectwise_boards[0]
        )
        with patch('djconnectwise.models.logger') as mock_logging:
            ticket.save()  # Should not log
            self.assertFalse(mock_logging.warning.called)
        with patch('djconnectwise.models.logger') as mock_logging:
            ticket.board = self.connectwise_boards[1]
            ticket.save()  # Should log
            self.assertTrue(mock_logging.warning.called)

    def test_save_calls_update_cw_when_kwarg_passed(self):
        board = self.connectwise_boards[0]
        ticket = Ticket.objects.create(
            summary='test',
            status=board.board_statuses.first(),
            board=board
        )
        with patch('djconnectwise.api.'
                   'ServiceAPIClient') as mock_serviceapiclient:
            instance = mock_serviceapiclient.return_value
            # Call save with no 'update_cw' kwarg- our mock should NOT
            # be called
            ticket.save()
            self.assertFalse(instance.update_ticket.called)
            # Now call it with 'update_cw'
            ticket.save(update_cw=True)
            self.assertTrue(instance.update_ticket.called)

    def test_update_cw(self):
        # Verify update_cw calls the API client
        board = self.connectwise_boards[0]
        ticket = Ticket.objects.create(
            summary='test',
            status=board.board_statuses.first(),
            board=board
        )
        with patch('djconnectwise.api.'
                   'ServiceAPIClient') as mock_serviceapiclient:
            instance = mock_serviceapiclient.return_value
            ticket.update_cw()
            self.assertTrue(instance.update_ticket.called)

    def test_close_ticket(self):
        # Verify close calls save.
        board = self.connectwise_boards[0]
        ticket = Ticket.objects.create(
            summary='test',
            status=board.board_statuses.first(),
            board=board
        )
        with patch.object(ticket, 'save') as mock_save:
            ticket.close()
            self.assertTrue(mock_save.called)

    def test_close_ticket_no_closed_statuses(self):
        # Raises an exception if there are no available closed statuses for
        # the ticket's board.
        board = self.connectwise_boards[0]
        board.board_statuses.filter(closed_status=True).delete()

        ticket = Ticket.objects.create(
            summary='test',
            status=board.board_statuses.first(),
            board=board
        )
        ticket.save()  # Should work
        with self.assertRaises(InvalidStatusError):
            ticket.close()

    @freeze_time("2018-08-23 17:24:34", tz_offset=-7)
    def test_calculate_sla_expiry(self):
        board = self.connectwise_boards[0]
        ticket = Ticket.objects.create(
            summary='test',
            status=board.board_statuses.get(name='New'),
            sla=Sla.objects.first(),
            priority=self.ticket_priorities[0],
            board=board,
            company=self.companies[0],
            entered_date_utc=timezone.now()
        )
        ticket.calculate_sla_expiry()

        self.assertEqual(
            str(ticket.sla_expire_date.astimezone(tz=None)),
            '2018-08-23 12:24:34-07:00'
            )
        self.assertEqual('respond', ticket.sla_stage)

    @freeze_time("2018-08-23 17:24:34", tz_offset=-7)
    def test_calculate_next_state_sla_expiry(self):
        board = self.connectwise_boards[0]
        old_status = board.board_statuses.get(name='New')
        ticket = Ticket.objects.create(
            summary='test',
            status=old_status,
            sla=Sla.objects.first(),
            priority=self.ticket_priorities[0],
            board=board,
            company=self.companies[0],
            entered_date_utc=timezone.now()
        )
        ticket.calculate_sla_expiry()

        ticket.status = board.board_statuses.get(name='Scheduled')
        ticket.calculate_sla_expiry()

        self.assertEqual(
            str(ticket.sla_expire_date.astimezone(tz=None)),
            '2018-08-24 09:24:34-07:00'
            )
        self.assertEqual('plan', ticket.sla_stage)

    @freeze_time("2018-08-23 23:24:34", tz_offset=-7)
    def test_calculate_sla_expiry_overnight(self):
        board = self.connectwise_boards[0]
        ticket = Ticket.objects.create(
            summary='test',
            status=board.board_statuses.get(name='New'),
            sla=Sla.objects.first(),
            priority=self.ticket_priorities[0],
            board=board,
            company=self.companies[0],
            entered_date_utc=timezone.now()
        )
        ticket.calculate_sla_expiry()

        self.assertEqual(
            str(ticket.sla_expire_date.astimezone(tz=None)),
            '2018-08-24 09:24:34-07:00'
            )

    @freeze_time("2018-08-23 17:24:34", tz_offset=-7)
    def test_calculate_sla_expiry_several_days(self):
        board = self.connectwise_boards[0]
        ticket = Ticket.objects.create(
            summary='test',
            status=board.board_statuses.get(name='In Progress'),
            sla=Sla.objects.first(),
            priority=self.ticket_priorities[0],
            board=board,
            company=self.companies[0],
            entered_date_utc=timezone.now()
        )
        ticket.calculate_sla_expiry()

        self.assertEqual(
            str(ticket.sla_expire_date.astimezone(tz=None)),
            '2018-09-06 16:24:34-07:00'
            )

    @freeze_time("2018-08-25 15:24:34", tz_offset=0)
    def test_calculate_sla_expiry_out_of_hours(self):
        board = self.connectwise_boards[0]
        ticket = Ticket.objects.create(
            summary='test',
            status=board.board_statuses.get(name='New'),
            sla=Sla.objects.first(),
            priority=self.ticket_priorities[0],
            board=board,
            company=self.companies[0],
            entered_date_utc=timezone.now()
        )
        ticket.calculate_sla_expiry()

        self.assertEqual(
            str(ticket.sla_expire_date.astimezone(tz=None)),
            '2018-08-27 10:00:34-07:00'
            )

    @freeze_time("2018-08-23 17:24:34", tz_offset=-7)
    def test_resolve_sla(self):
        # Test entering a resolved state
        board = self.connectwise_boards[0]
        old_status = board.board_statuses.get(name='New')
        ticket = Ticket.objects.create(
            summary='test',
            status=old_status,
            sla=Sla.objects.first(),
            priority=self.ticket_priorities[0],
            board=board,
            company=self.companies[0],
            entered_date_utc=timezone.now()
        )

        ticket.status = board.board_statuses.get(name='Completed')
        ticket.calculate_sla_expiry()

        self.assertFalse(ticket.sla_expire_date)
        self.assertEqual(ticket.sla_stage, 'resolved')

    @freeze_time("2018-08-23 17:24:34", tz_offset=-7)
    def test_sla_enter_waiting(self):
        board = self.connectwise_boards[0]
        old_status = board.board_statuses.get(name='New')
        ticket = Ticket.objects.create(
            summary='test',
            status=old_status,
            sla=Sla.objects.first(),
            priority=self.ticket_priorities[0],
            board=board,
            company=self.companies[0],
            entered_date_utc=timezone.now()
        )

        ticket.status = board.board_statuses.get(name='Waiting For Client')
        ticket.calculate_sla_expiry()

        self.assertFalse(ticket.sla_expire_date)
        self.assertEqual(ticket.sla_stage, 'waiting')
        self.assertEqual(
            timezone.now(),
            ticket.do_not_escalate_date
            )

    @freeze_time("2018-08-23 17:24:34", tz_offset=-7)
    def test_sla_exit_waiting(self):
        board = self.connectwise_boards[0]
        entered_date_utc = timezone.now() - datetime.timedelta(hours=3)
        waiting_status = board.board_statuses.get(name='Waiting For Client')
        do_not_escalate_date = timezone.now()-datetime.timedelta(hours=1)

        ticket = Ticket.objects.create(
            summary='test',
            status=waiting_status,
            sla=Sla.objects.first(),
            priority=self.ticket_priorities[0],
            board=board,
            company=self.companies[0],
            entered_date_utc=entered_date_utc,
        )
        ticket.do_not_escalate_date = do_not_escalate_date
        ticket.status = board.board_statuses.get(name='New')
        ticket.save()

        self.assertEqual(ticket.sla_stage, 'respond')
        self.assertEqual(ticket.minutes_waiting, 60)
        self.assertFalse(ticket.do_not_escalate_date)
        self.assertTrue(ticket.sla_expire_date)

    def test_sla_exit_waiting_out_of_hours(self):
        board = self.connectwise_boards[0]
        entered_date_utc = timezone.now()
        waiting_status = board.board_statuses.get(name='Waiting For Client')
        ticket = Ticket.objects.create(
            summary='test',
            status=waiting_status,
            sla=Sla.objects.first(),
            priority=self.ticket_priorities[0],
            board=board,
            company=self.companies[0],
            entered_date_utc=entered_date_utc,
            do_not_escalate_date=timezone.now()
        )

        ticket.status = board.board_statuses.get(name='New')
        ticket.calculate_sla_expiry()

        self.assertEqual(ticket.sla_stage, 'respond')
        self.assertFalse(ticket.do_not_escalate_date)
        self.assertTrue(ticket.sla_expire_date)

    @freeze_time("2018-08-25 17:24:34", tz_offset=-7)
    def test_sla_exit_waiting_weekend(self):
        board = self.connectwise_boards[0]
        entered_date_utc = timezone.now() - datetime.timedelta(hours=4)
        waiting_status = board.board_statuses.get(name='Waiting For Client')
        ticket = Ticket.objects.create(
            summary='test',
            status=waiting_status,
            sla=Sla.objects.first(),
            priority=self.ticket_priorities[0],
            board=board,
            company=self.companies[0],
            entered_date_utc=entered_date_utc,
            do_not_escalate_date=timezone.now() - datetime.timedelta(hours=2)
        )

        ticket.status = board.board_statuses.get(name='New')
        ticket.calculate_sla_expiry()

        self.assertEqual(ticket.sla_stage, 'respond')
        self.assertFalse(ticket.do_not_escalate_date)
        self.assertTrue(ticket.sla_expire_date)
        self.assertEqual(ticket.minutes_waiting, 0)

    def test_sla_exit_waiting_to_lower_status(self):
        board = self.connectwise_boards[0]
        test_date = timezone.now()
        waiting_status = board.board_statuses.get(name='Waiting For Client')
        ticket = Ticket.objects.create(
            summary='test',
            status=waiting_status,
            sla=Sla.objects.first(),
            priority=self.ticket_priorities[0],
            board=board,
            company=self.companies[0],
            entered_date_utc=test_date,
            do_not_escalate_date=timezone.now(),
            date_responded_utc=test_date
        )

        ticket.status = board.board_statuses.get(name='New')
        ticket.calculate_sla_expiry()
        self.assertEqual(ticket.sla_stage, 'plan')

    def test_sla_exit_resolved_only_to_resolve(self):
        board = self.connectwise_boards[0]
        test_date = timezone.now()
        old_status = board.board_statuses.get(name='Completed')
        ticket = Ticket.objects.create(
            summary='test',
            status=old_status,
            sla=Sla.objects.first(),
            priority=self.ticket_priorities[0],
            board=board,
            company=self.companies[0],
            entered_date_utc=test_date,
            do_not_escalate_date=timezone.now(),
            date_resolved_utc=test_date
        )

        ticket.status = board.board_statuses.get(name='In Progress')
        ticket.calculate_sla_expiry()

        self.assertEqual(ticket.sla_stage, 'resolve')

    def test_lowest_possible_stage(self):
        board = self.connectwise_boards[0]
        ticket = Ticket.objects.create(
            summary='test',
            status=board.board_statuses.get(name='Waiting For Client'),
            sla=Sla.objects.first(),
            priority=self.ticket_priorities[0],
            board=board,
            company=self.companies[0],
            entered_date_utc=timezone.now(),
            do_not_escalate_date=timezone.now()
        )

        result = ticket.sla_state._lowest_possible_stage('resolved')
        self.assertEqual('resolved', result)

        result = ticket.sla_state._lowest_possible_stage('respond')
        self.assertEqual('respond', result)

        result = ticket.sla_state._lowest_possible_stage('plan')
        self.assertEqual('plan', result)

        result = ticket.sla_state._lowest_possible_stage('resolve')
        self.assertEqual('resolve', result)

        ticket.date_responded_utc = timezone.now()
        result = ticket.sla_state._lowest_possible_stage('respond')
        self.assertEqual('plan', result)

        ticket.date_resplan_utc = timezone.now()
        result = ticket.sla_state._lowest_possible_stage('respond')
        self.assertEqual('resolve', result)

    @freeze_time("2021-08-23 08:24:34", tz_offset=-7)
    def test_calculate_sla_on_priority_change(self):
        """
        Verify that SLA expiry dates are re-calculated when only
        the SLA priority is changed.
        """
        board = self.connectwise_boards[0]
        ticket = Ticket.objects.create(
            summary='test',
            status=board.board_statuses.get(name='New'),
            sla=Sla.objects.first(),
            priority=self.ticket_priorities[2],
            board=board,
            company=self.companies[0],
            entered_date_utc=timezone.now()
        )

        # Time in UTC-7 is 2021-08-23 01:24:34
        self.assertEqual(
            str(ticket.sla_expire_date.astimezone(tz=None)),
            '2021-08-23 07:24:34-07:00'
        )
        self.assertEqual('respond', ticket.sla_stage)

        # Change to a priority with less response hours
        ticket.priority = self.ticket_priorities[1]
        ticket.save()
        self.assertEqual(
            str(ticket.sla_expire_date.astimezone(tz=None)),
            '2021-08-23 03:24:34-07:00'
        )

    @freeze_time("2021-08-23 00:24:34", tz_offset=-7)
    def test_calculate_sla_on_status_and_priority_change(self):
        """
        Verify that SLA expiry dates are re-calculated correctly when priority
        and status are changed.
        """
        board = self.connectwise_boards[0]
        ticket = Ticket.objects.create(
            summary='test',
            status=board.board_statuses.get(name='New'),
            sla=Sla.objects.first(),
            priority=self.ticket_priorities[1],
            board=board,
            company=self.companies[0],
            entered_date_utc=timezone.now()
        )

        # Time in UTC-7 is 2021-08-23 17:24:34
        self.assertEqual(
            str(ticket.sla_expire_date.astimezone(tz=None)),
            '2021-08-23 10:00:34-07:00'
        )
        self.assertEqual('respond', ticket.sla_stage)

        ticket.status = board.board_statuses.get(name='Scheduled')
        ticket.save()

        self.assertEqual('plan', ticket.sla_stage)
        self.assertEqual(
            str(ticket.sla_expire_date.astimezone(tz=None)),
            '2021-08-23 12:00:34-07:00'
        )

        ticket.priority = self.ticket_priorities[2]
        ticket.save()
        self.assertEqual(
            str(ticket.sla_expire_date.astimezone(tz=None)),
            '2021-08-24 11:00:34-07:00'
        )



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations, models
import django_extensions.db.fields


class Migration(migrations.Migration):

    dependencies = [
        ('djconnectwise', '0042_auto_20171115_1131'),
    ]

    operations = [
        migrations.CreateModel(
            name='ProjectStatus',
            fields=[
                ('id', models.AutoField(verbose_name='ID', primary_key=True, auto_created=True, serialize=False)),
                ('created', django_extensions.db.fields.CreationDateTimeField(auto_now_add=True, verbose_name='created')),
                ('modified', django_extensions.db.fields.ModificationDateTimeField(auto_now=True, verbose_name='modified')),
                ('name', models.CharField(max_length=30)),
                ('default_flag', models.BooleanField(default=False)),
                ('inactive_flag', models.BooleanField(default=False)),
                ('closed_flag', models.BooleanField(default=False)),
            ],
            options={
                'verbose_name_plural': 'Project statuses',
                'ordering': ('name',),
            },
        ),
    ]



import logging
from djconnectwise.api import SystemAPIClient
from djconnectwise.utils import DjconnectwiseSettings


logger = logging.getLogger(__name__)
# We need to have exactly these callbacks. Fields that are omitted are not
# considered. (URL is set at runtime.)
# See https://developer.connectwise.com/Products/Manage/Developer_Guide#Callbacks_(Webhooks)  # noqa
NEEDED_CALLBACKS = [
  {
    "type": "ticket",
    "description": "Kanban application ticket callback",
    "url": None,
    "objectId": 1,
    "level": "owner",
  },
  {
    "type": "project",
    "description": "Kanban application project callback",
    "url": None,
    "objectId": 1,
    "level": "owner",
  },
  {
    "type": "company",
    "description": "Kanban application company callback",
    "url": None,
    "objectId": 1,
    "level": "owner",
  },
  {
    "type": "opportunity",
    "description": "Kanban application opportunity callback",
    "url": None,
    "objectId": 1,
    "level": "owner",
  },
  {
    "type": "activity",
    "description": "Kanban application activity callback",
    "url": None,
    "objectId": 1,
    "level": "owner",
  },
  {
    "type": "schedule",
    "description": "Kanban application schedule entry callback",
    "url": None,
    "objectId": 1,
    "level": "owner",
  }
]


class CallbacksHandler:
    def __init__(self):
        super().__init__()
        self.system_client = SystemAPIClient()
        self.settings = DjconnectwiseSettings().get_settings()

    def get_callbacks(self):
        results = []
        page = 1
        while True:
            page_records = self.system_client.get_callbacks(
                page=page,
                page_size=self.settings['batch_size'],
                conditions=[
                    'url contains "{}"'.format(self.settings['callback_host'])
                ]
            )
            results += page_records
            page += 1
            if len(page_records) < self.settings['batch_size']:
                # This page wasn't full, so there's no more records after
                # this page.
                break
        return results

    def get_needed_callbacks(self):
        """
        Return a list of callbacks
        """
        result = NEEDED_CALLBACKS.copy()
        for cb in result:
            cb['url'] = '{}{}'.format(
                self.settings['callback_host'],
                self.settings['callback_url']
            )
        return result

    def _calculate_missing_unneeded_callbacks(self,
                                              needed_callbacks,
                                              current_callbacks):
        """
        Given the list of needed callbacks and current callbacks, figure out
        which callbacks need to be registered and which ones need to be
        removed.

        Returns a tuple of (need-to-add, need-to-remove) callbacks.
        """
        # For each current callback that matches a needed callback, delete it
        # from both needed and current.
        current_indexes_delete = set()
        needed_indexes_delete = set()
        field_names = ['type', 'description', 'url', 'objectId', 'level']
        # Add the index of the element to delete to a list, because it's
        # an error to delete elements out of a list while iterating through.
        for i_cur, current in enumerate(current_callbacks):
            for i_need, needed in enumerate(needed_callbacks):
                all_matched = True
                for field in field_names:
                    if current[field] != needed[field]:
                        # If any field doesn't match, don't delete the cb
                        all_matched = False
                if all_matched:
                    current_indexes_delete.add(i_cur)
                    needed_indexes_delete.add(i_need)
        for i_cur in reversed(sorted(current_indexes_delete)):
            # Delete from the end
            current_callbacks.pop(i_cur)
        for i_need in reversed(sorted(needed_indexes_delete)):
            # Delete from the end
            needed_callbacks.pop(i_need)

        return needed_callbacks, current_callbacks

    def ensure_registered(self):
        """
        Do the minimum changes to ensure our callbacks are registered
        exactly once.
        """
        needed_callbacks = self.get_needed_callbacks()
        current_callbacks = self.get_callbacks()
        callbacks_to_add, callbacks_to_remove = \
            self._calculate_missing_unneeded_callbacks(
                needed_callbacks, current_callbacks
            )

        for callback in callbacks_to_add:
            self._register_callback(callback)
        for callback in callbacks_to_remove:
            self._delete_callback(callback)

    def ensure_deleted(self):
        """Do the needful to ensure our callbacks are gone."""
        callbacks = self.get_callbacks()
        for callback in callbacks:
            self._delete_callback(callback)

    def _register_callback(self, callback):
        logger.info('Registering {} callback'.format(callback['type']))
        self.system_client.create_callback(callback)

    def _delete_callback(self, callback):
        logger.info('Deleting callback {}'.format(callback['id']))
        self.system_client.delete_callback(callback['id'])



# Generated by Django 2.0 on 2018-08-15 11:16

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('djconnectwise', '0068_merge_20180814_1104'),
    ]

    operations = [
        migrations.AlterField(
            model_name='timeentry',
            name='billable_option',
            field=models.CharField(choices=[('Billable', 'Billable'), ('DoNotBill', 'Do Not Bill'), ('NoCharge', 'No Charge'), ('NoDefault', 'No Default')], max_length=250),
        ),
        migrations.AlterField(
            model_name='timeentry',
            name='charge_to_type',
            field=models.CharField(choices=[('ServiceTicket', 'Service Ticket'), ('ProjectTicket', 'Project Ticket'), ('ChargeCode', 'Charge Code'), ('Activity', 'Activity')], max_length=250),
        ),
    ]



from collections import OrderedDict

from djconnectwise import sync, api
from djconnectwise.utils import DjconnectwiseSettings

from django.core.management.base import BaseCommand, CommandError
from django.utils.translation import gettext_lazy as _

OPTION_NAME = 'connectwise_object'


class Command(BaseCommand):
    help = str(_('Synchronize the specified object with the Connectwise API'))

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # This can be replaced with a single instantiation of an OrderedDict
        # using kwargs in Python 3.6. But we need Python 3.5 compatibility for
        # now.
        # See https://www.python.org/dev/peps/pep-0468/.
        synchronizers = (
            ('member', sync.MemberSynchronizer, _('Member')),
            ('board', sync.BoardSynchronizer, _('Board')),
            ('team', sync.TeamSynchronizer, _('Team')),
            ('board_status', sync.BoardStatusSynchronizer, _('Board Status')),
            ('priority', sync.PrioritySynchronizer, _('Priority')),
            ('project_status', sync.ProjectStatusSynchronizer,
                _('Project Status')),
            ('project_type', sync.ProjectTypeSynchronizer,
                _('Project Type')),
            ('project', sync.ProjectSynchronizer, _('Project')),
            ('project_phase', sync.ProjectPhaseSynchronizer,
                _('Project Phase')),
            ('territory', sync.TerritorySynchronizer,
                _('Territory')),
            ('company_status', sync.CompanyStatusSynchronizer,
                _('Company Status')),
            ('company_type', sync.CompanyTypeSynchronizer, _('Company Type')),
            ('company', sync.CompanySynchronizer, _('Company')),
            ('location', sync.LocationSynchronizer, _('Location')),
            ('opportunity_status', sync.OpportunityStatusSynchronizer,
                _('Opportunity Status')),
            ('opportunity_stage', sync.OpportunityStageSynchronizer,
                _('Opportunity Stage')),
            ('opportunity_type', sync.OpportunityTypeSynchronizer,
                _('Opportunity Type')),
            ('sales_probability', sync.SalesProbabilitySynchronizer,
                _('Sales Probability')),
            ('opportunity', sync.OpportunitySynchronizer,
                _('Opportunity')),
            ('holiday_list', sync.HolidayListSynchronizer,
                _('Holiday List')),
            ('holiday', sync.HolidaySynchronizer,
                _('Holiday')),
            ('calendar', sync.CalendarSynchronizer,
                _('Calendar')),
            ('company_other', sync.MyCompanyOtherSynchronizer,
                _('Company Other')),
            ('sla', sync.SLASynchronizer,
                _('Sla')),
            ('sla_priority', sync.SLAPrioritySynchronizer,
                _('Sla Priority')),
            ('type', sync.TypeSynchronizer,
                _('Type')),
            ('sub_type', sync.SubTypeSynchronizer,
                _('Sub Type')),
            ('item', sync.ItemSynchronizer,
                _('Item')),
            ('ticket', sync.ServiceTicketSynchronizer, _('Ticket')),
            ('project_ticket', sync.ProjectTicketSynchronizer,
                _('Project Ticket')),
            ('work_type', sync.WorkTypeSynchronizer,
                _('Work Type')),
            ('work_role', sync.WorkRoleSynchronizer,
                _('Work Role')),
            ('agreement', sync.AgreementSynchronizer,
                _('Agreement')),
            ('activity_status', sync.ActivityStatusSynchronizer,
                _('Activity Status')),
            ('activity_type', sync.ActivityTypeSynchronizer,
                _('Activity Type')),
            ('activity', sync.ActivitySynchronizer,
                _('Activity')),
            ('schedule_type', sync.ScheduleTypeSynchronizer,
                _('Schedule Type')),
            ('schedule_status', sync.ScheduleStatusSynchronizer,
                _('Schedule Status')),
            ('schedule_entry', sync.ScheduleEntriesSynchronizer,
                _('Schedule Entry')),
            ('project_team_member', sync.ProjectTeamMemberSynchronizer,
                _('Project Team Member')),
            ('ticket_udf', sync.TicketUDFSynchronizer, _('Ticket UDF')),
            ('project_udf', sync.ProjectUDFSynchronizer, _('Project UDF')),
            ('activity_udf', sync.ActivityUDFSynchronizer, _('Activity UDF')),
            ('opportunity_udf', sync.OpportunityUDFSynchronizer,
                _('Opportunity UDF')),
        )

        settings = DjconnectwiseSettings().get_settings()
        if settings['sync_time_and_note_entries']:
            synchronizers = synchronizers + (
                ('service_note', sync.ServiceNoteSynchronizer,
                    _('Service Note')),
                ('opportunity_note', sync.OpportunityNoteSynchronizer,
                    _('Opportunity Note')),
                ('time_entry', sync.TimeEntrySynchronizer,
                    _('Time Entry'))
            )

        self.synchronizer_map = OrderedDict()
        for name, synchronizer, obj_name in synchronizers:
            self.synchronizer_map[name] = (synchronizer, obj_name)

    def add_arguments(self, parser):
        parser.add_argument(OPTION_NAME, nargs='?', type=str)
        parser.add_argument('--full',
                            action='store_true',
                            dest='full',
                            default=False)

    def sync_by_class(self, sync_class, obj_name, full_option=False):
        synchronizer = sync_class(full=full_option)

        created_count, updated_count, skipped_count, deleted_count = \
            synchronizer.sync()

        msg = _('{} Sync Summary - Created: {}, Updated: {}, Skipped: {}')
        fmt_msg = msg.format(obj_name, created_count, updated_count,
                             skipped_count)

        if full_option:
            msg = _('{} Sync Summary - Created: {}, Updated: {}, '
                    'Skipped: {}, Deleted: {}')
            fmt_msg = msg.format(obj_name, created_count, updated_count,
                                 skipped_count, deleted_count)

        self.stdout.write(fmt_msg)

    def handle(self, *args, **options):
        sync_classes = []
        connectwise_object_arg = options[OPTION_NAME]
        full_option = options.get('full', False)

        if connectwise_object_arg:
            object_arg = connectwise_object_arg
            sync_tuple = self.synchronizer_map.get(object_arg)

            if sync_tuple:
                sync_classes.append(sync_tuple)
            else:
                msg = _('Invalid CW object {}, '
                        'choose one of the following: \n{}')
                options_txt = ', '.join(self.synchronizer_map.keys())
                msg = msg.format(sync_tuple, options_txt)
                raise CommandError(msg)
        else:
            sync_classes = self.synchronizer_map.values()

        failed_classes = 0
        error_messages = ''

        for sync_class, obj_name in sync_classes:
            try:
                self.sync_by_class(sync_class, obj_name,
                                   full_option=full_option)

            except api.ConnectWiseAPIError as e:
                msg = 'Failed to sync {}: {}'.format(obj_name, e)
                self.stderr.write(msg)
                error_messages += '{}\n'.format(msg)
                failed_classes += 1

        if failed_classes > 0:
            msg = '{} class{} failed to sync.\n'.format(
                failed_classes,
                '' if failed_classes == 1 else 'es',
            )
            msg += 'Errors:\n'
            msg += error_messages
            raise CommandError(msg)



# Generated by Django 3.1.1 on 2020-10-05 10:54

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('djconnectwise', '0128_auto_20201003_1140'),
    ]

    operations = [
        migrations.AlterField(
            model_name='ticket',
            name='approved',
            field=models.BooleanField(null=True),
        ),
    ]



# Generated by Django 2.1 on 2019-07-18 14:06

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('djconnectwise', '0097_merge_20190718_1325'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='ticket',
            name='agreement_id',
        ),
        migrations.AddField(
            model_name='ticket',
            name='agreement',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='agreement_tickets', to='djconnectwise.Agreement'),
        ),
    ]



# Generated by Django 3.1.2 on 2020-10-15 16:14

from django.db import migrations, models
import django_extensions.db.fields


class Migration(migrations.Migration):

    dependencies = [
        ('djconnectwise', '0131_ticket_udf'),
    ]

    operations = [
        migrations.CreateModel(
            name='ActivityUDF',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', django_extensions.db.fields.CreationDateTimeField(auto_now_add=True, verbose_name='created')),
                ('modified', django_extensions.db.fields.ModificationDateTimeField(auto_now=True, verbose_name='modified')),
                ('caption', models.CharField(blank=True, max_length=50, null=True)),
                ('type', models.CharField(blank=True, max_length=50, null=True)),
                ('entryMethod', models.CharField(blank=True, max_length=50, null=True)),
                ('numberOfDecimals', models.PositiveSmallIntegerField(blank=True, null=True)),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='OpportunityUDF',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', django_extensions.db.fields.CreationDateTimeField(auto_now_add=True, verbose_name='created')),
                ('modified', django_extensions.db.fields.ModificationDateTimeField(auto_now=True, verbose_name='modified')),
                ('caption', models.CharField(blank=True, max_length=50, null=True)),
                ('type', models.CharField(blank=True, max_length=50, null=True)),
                ('entryMethod', models.CharField(blank=True, max_length=50, null=True)),
                ('numberOfDecimals', models.PositiveSmallIntegerField(blank=True, null=True)),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='ProjectUDF',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', django_extensions.db.fields.CreationDateTimeField(auto_now_add=True, verbose_name='created')),
                ('modified', django_extensions.db.fields.ModificationDateTimeField(auto_now=True, verbose_name='modified')),
                ('caption', models.CharField(blank=True, max_length=50, null=True)),
                ('type', models.CharField(blank=True, max_length=50, null=True)),
                ('entryMethod', models.CharField(blank=True, max_length=50, null=True)),
                ('numberOfDecimals', models.PositiveSmallIntegerField(blank=True, null=True)),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='TicketUDF',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', django_extensions.db.fields.CreationDateTimeField(auto_now_add=True, verbose_name='created')),
                ('modified', django_extensions.db.fields.ModificationDateTimeField(auto_now=True, verbose_name='modified')),
                ('caption', models.CharField(blank=True, max_length=50, null=True)),
                ('type', models.CharField(blank=True, max_length=50, null=True)),
                ('entryMethod', models.CharField(blank=True, max_length=50, null=True)),
                ('numberOfDecimals', models.PositiveSmallIntegerField(blank=True, null=True)),
            ],
            options={
                'abstract': False,
            },
        ),
    ]



from django.apps import AppConfig


class DjangoConnectwiseConfig(AppConfig):
    name = 'djconnectwise'

    def ready(self):
        # Register signal handlers using decorators.
        from djconnectwise import signals
